"""Functions for enumerations or correlations.

Example, enumerating USB activity on a sensor from a programatic
analysis of registry modifications.
"""
import os
import logging

from cbapi.response import CbResponseAPI

from cbinterface.helpers import as_configured_timezone


LOGGER = logging.getLogger("cbinterface.response.enumerations")


def logon_history(cb: CbResponseAPI, hostname_or_username_query) -> None:
    """Given hostname or username, enumerate logon history.

    Note, this is an analysis of the WINDOWS behavior when a new user
    session is started. It's informative for analysts, not authoritative.
    It will NOT show processes ran under other users, as often is the case
    with enterprise admin activity.
    """
    from cbinterface.response.query import make_process_query

    if not (hostname_or_username_query.startswith("hostname:") or hostname_or_username_query.startswith("username:")):
        LOGGER.error(f"Must supply 'hostname:' or 'username:' field and value.")
        return False

    query = f"process_name:userinit.exe parent_name:winlogon.exe {hostname_or_username_query}"

    processes = make_process_query(cb, query)
    if processes and len(processes) > 0:
        timezone_string = os.environ.get("CBINTERFACE_TIMEZONE", "GMT")
        print(f"\n\t{timezone_string} Time    \t|\tUsername\t|\tHostname")
        for proc in processes:
            start_time = as_configured_timezone(proc.start, apply_time_format="%Y-%m-%d %H:%M:%S%z")
            print("  {}\t    {}\t\t{}".format(start_time, proc.username, proc.hostname))
        print()
    return



"""All things LR sessions."""

import io
import os
import time
import logging
import threading

from cbapi.response.cblr import LiveResponseSession, LiveResponseSessionManager
from cbapi.response import CbResponseAPI, Sensor
from cbapi.errors import ObjectNotFoundError, TimeoutError
from cbapi.live_response_api import CbLRManagerBase, LiveResponseJobScheduler, WorkItem, poll_status

from typing import List, Union

from cbinterface.commands import BaseSessionCommand
from cbinterface.response.sensor import is_sensor_online

LOGGER = logging.getLogger("cbinterface.response.session")

CBLR_BASE = "/api/v1/cblr"


class CustomLiveResponseSessionManager(LiveResponseSessionManager):
    def __init__(self, cb, timeout=30, custom_session_keepalive=False):
        super().__init__(cb, timeout=timeout, keepalive_sessions=False)
        # NOTE: keepalives will automatically close any sessions NOT in an active state.
        # So, don't send keepalives to pending sessions or the session will close.
        # For this reason, don't start sessions on sensor until those sensors are online.
        # Still, if a sensor is slow to check in a timing hicup can occur. The result being, a
        # keepalive will close a pending session that would have otherwise have become active.

        if custom_session_keepalive:
            self._cleanup_thread = threading.Thread(target=self._keep_active_sessions_alive_thread)
            self._cleanup_thread.daemon = True
            self._cleanup_thread.start()

        # for storing initiatied commands
        self.commands = []

    def active_session(self, sensor: Sensor):
        """Return any active sessions on this sensor or None."""
        active = sensor_live_response_sessions(sensor, active_or_pending=True)
        if active:
            session_data = active[0]
            session_id = session_data["id"]
            LOGGER.info(f"found existing session id={session_id} in '{session_data['status']}' state")
            self._sessions[sensor.id] = self.cblr_session_cls(self, session_id, sensor.id, session_data=session_data)
            return self._sessions[sensor.id]
        return None

    def get_session(self, sensor: Sensor):
        """Get or create LR session."""
        active_session = self.active_session(sensor)
        if isinstance(active_session, self.cblr_session_cls):
            return active_session

        session_data = self._cb.post_object(f"{CBLR_BASE}/session", {"sensor_id": sensor.id}).json()
        session_id = session_data["id"]
        LOGGER.info(f"created session id={session_id}")
        self._sessions[sensor.id] = self.cblr_session_cls(self, session_id, sensor.id, session_data=session_data)
        return self._sessions[sensor.id]

    def wait_for_active_session(self, sensor: Sensor, timeout=86400):
        """Return active session or None.

        Default timeout is 7 days.
        """
        LOGGER.info(
            f"attempting to get active session on sensor {sensor.id} (hostname:{sensor.hostname}) for up to {timeout/60} minutes"
        )
        start_time = time.time()
        session = None
        status = None
        while status != "active" and time.time() - start_time < timeout:
            if not is_sensor_online(sensor):
                LOGGER.debug(f"waiting for sensor {sensor.id} to come online...")
                time.sleep(1)
                continue
            if status is None:
                session = self.get_session(sensor)
            status = get_session_status(self._cb, session.session_id)
            time.sleep(0.5)

        if session and is_session_active(session):
            LOGGER.info(f"got active session {session.session_id} on sensor {sensor.id}.")
        return session

    def submit_command(self, command: BaseSessionCommand, sensor: Union[int, Sensor]):
        """
        Create a new job to be executed as a Live Response.

        Args:
            command (BaseSessionCommand): The job to be scheduled.
            sensor (Sensor): Sensor to execute job on.
        Returns:
            Future: A reference to the running job.
        """
        assert isinstance(command, BaseSessionCommand)

        sensor_id = sensor
        if isinstance(sensor, Sensor):
            sensor_id = sensor.id

        if self._job_scheduler is None:
            # spawn the scheduler thread
            self._job_scheduler = LiveResponseJobScheduler(self._cb)
            self._job_scheduler.start()

        if sensor_id not in self._sessions:
            sensor = Sensor(self._cb, sensor_id, force_init=True)
            active_session = self.active_session(sensor)
            if active_session is None:
                self.wait_for_active_session(sensor)

        work_item = WorkItem(command.run, sensor_id)
        self._job_scheduler.submit_job(work_item)
        command.future = work_item.future
        command._sensor_id = sensor_id
        command.session_id = self._sessions[sensor_id].session_id
        command.session_data = self._sessions[sensor_id].session_data
        self.commands.append(command)
        return command

    def yield_completed_commands(self):
        """Wait for commands to complete, process results.

        Monitor commands and sessions.
        """
        LOGGER.info(f"waiting for {len(self.commands)} commands to complete ...")
        while self.commands:
            for cmd in self.commands:
                if not cmd.initiatied:
                    LOGGER.error(f"skipping uninitialized command: {cmd}")
                    self.commands.remove(cmd)
                    continue
                if cmd.exception:
                    LOGGER.error(f"exception for {cmd}: {cmd.exception}")
                    self.commands.remove(cmd)
                    continue
                if not get_session_by_id(self._cb, cmd.session_id):
                    LOGGER.error(f"session {cmd.session_id} is gone. command has gone to the void: {cmd}")
                    self.commands.remove(cmd)
                    continue

                if cmd.has_result:
                    LOGGER.debug(f"yielding {cmd}")
                    yield cmd
                    self.commands.remove(cmd)

                # yield time for completion
                time.sleep(0.7)

    def process_completed_commands(self):
        for cmd in self.yield_completed_commands():
            LOGGER.debug(f"processing => {cmd}")
            cmd.process_result()

    def _keep_active_sessions_alive_thread(self):
        """Used by a thread to ping active sessions so they don't
        close on long running session commands.
        """
        LOGGER.debug("Starting custom Live Response session keepalive and cleanup task")
        while True:
            time.sleep(self._timeout)

            delete_list = []
            with self._session_lock:
                for session in self._sessions.values():
                    if session._refcount == 0:
                        delete_list.append(session.sensor_id)
                    else:
                        try:
                            if is_session_active(session):
                                LOGGER.info(f"sending keepalive for session {session.session_id}")
                                self._send_keepalive(session.session_id)
                        except ObjectNotFoundError:
                            LOGGER.debug(
                                f"Session {session.session_id} for sensor {session.sensor_id} not valid any longer, removing from cache"
                            )
                            delete_list.append(session.sensor_id)
                        except Exception as e:
                            LOGGER.warning(
                                f"Keepalive on session {session.session_id} (sensor {session.sensor_id}) failed with unknown error: {e}"
                            )
                            delete_list.append(session.sensor_id)

                for sensor_id in delete_list:
                    self._close_session(self._sessions[sensor_id].session_id)
                    del self._sessions[sensor_id]


def all_live_response_sessions(cb: CbResponseAPI) -> List:
    """List all LR sessions still in server memory."""
    return [sesh for sesh in cb.get_object(f"{CBLR_BASE}/session")]


def active_live_response_sessions(cb: CbResponseAPI) -> List:
    """Return active LR sessions."""
    return [sesh for sesh in cb.get_object(f"{CBLR_BASE}/session?active_only=true")]


def sensor_live_response_sessions(sensor: Sensor, active_or_pending=False):
    """Get sessions associated to this sensor."""
    sessions = [session for session in all_live_response_sessions(sensor._cb) if session["sensor_id"] == sensor.id]
    if active_or_pending:
        return [session for session in sessions if session["status"] == "active" or session["status"] == "pending"]
    return sessions


def sensor_live_response_sessions_by_sensor_id(cb: CbResponseAPI, sensor_id: Union[int, str]):
    """Get sessions associated to this sensor by sensor id."""
    if isinstance(sensor_id, str):
        sensor_id = int(sensor_id)
    sessions = [session for session in all_live_response_sessions(cb) if session["sensor_id"] == sensor_id]
    return sessions


def get_session_by_id(cb: CbResponseAPI, session_id):
    """Get a LR session object by id."""
    try:
        return cb.get_object(f"{CBLR_BASE}/session/{session_id}")
    except ObjectNotFoundError:
        LOGGER.warning(f"no live resonse session by ID={session_id}")
        return None


def get_session_status(cb: CbResponseAPI, session_id):
    """Return any session status or None."""
    session = get_session_by_id(cb, session_id)
    if session is None:
        return None
    return session["status"]


def is_session_active(session: LiveResponseSession):
    """Return True if session is active."""
    session_data = get_session_by_id(session._cb, session.session_id)
    if session_data is None:
        return None
    if session_data["status"] == "active":
        return True
    return False


def get_session_commands(cb: CbResponseAPI, session_id: str):
    """List commands for this session."""
    try:
        return cb.get_object(f"{CBLR_BASE}/session/{session_id}/command")
    except ObjectNotFoundError:
        LOGGER.warning(f"no live resonse session by ID={session_id}")
        return None


def get_command_result(cb: CbResponseAPI, session_id: str, command_id: str):
    """Get results of a LR session command."""
    try:
        return cb.get_object(f"{CBLR_BASE}/session/{session_id}/command/{command_id}")
    except ObjectNotFoundError:
        LOGGER.warning(f"no live resonse session and/or command combination for {session_id}:{command_id}")
        return None


def get_file_content(cb: CbResponseAPI, session_id: str, file_id: str):
    """Get file content stored in LR session and write the file locally."""
    from cbinterface.helpers import get_os_independent_filepath

    try:
        file_metadata = cb.get_object(f"{CBLR_BASE}/session/{session_id}/file/{file_id}")
        if file_metadata:
            filepath = get_os_independent_filepath(file_metadata["file_name"])
            filename = f"{session_id}_{filepath.name}"
        result = cb.session.get(f"{CBLR_BASE}/session/{session_id}/file/{file_id}/content", stream=True)
        if result.status_code != 200:
            LOGGER.error(
                f"got {result.status_code} from server getting file {file_id} content for session {session_id}"
            )
            return
        with open(filename, "wb") as fp:
            for chunk in result.iter_content(io.DEFAULT_BUFFER_SIZE):
                fp.write(chunk)
        if os.path.exists(filename):
            LOGGER.info(f"wrote: {filename}")
        return os.path.exists(filename)
    except ObjectNotFoundError:
        LOGGER.warning(f"no file {file_id} content with session {session_id}")
        return



"""Functions for enumerations or correlations.

Example, enumerating USB activity on a sensor from a programatic
analysis of registry modifications.
"""
import os
import logging

from cbapi.psc.rest_api import CbPSCBaseAPI

from cbinterface.helpers import as_configured_timezone


LOGGER = logging.getLogger("cbinterface.psc.enumerations")


def search_for_usb_devices(cb: CbPSCBaseAPI, query):
    # XXX - This doesn't look implemented in TH?
    uri = f"/device_control/v3/orgs/{cb.credentials.org_key}/devices/_search"
    data = {"query": query}
    return cb.post_object(uri, data).json()


def logon_history(cb: CbPSCBaseAPI, hostname_or_username_query) -> None:
    """Given hostname or username, enumerate logon history.

    Note, this is an analysis of the WINDOWS behavior when a new user
    session is started. It's informative for analysts, not authoritative.
    It will NOT show processes ran under other users, as often is the case
    with enterprise admin activity.
    """
    from cbinterface.psc.query import make_process_query

    if not (
        hostname_or_username_query.startswith("device_name:")
        or hostname_or_username_query.startswith("process_username:")
    ):
        LOGGER.info(f"use 'device_name:' or 'process_username:' field to narrow enumeration search.")

    # query = f"process_name:userinit.exe parent_name:winlogon.exe {hostname_or_username_query}"
    # XXX will catch more than "log on"
    query = f"process_name:explorer.exe {hostname_or_username_query}"

    processes = make_process_query(cb, query)
    if processes and len(processes) > 0:
        timezone_string = os.environ.get("CBINTERFACE_TIMEZONE", "GMT")
        print(f"\n\t{timezone_string} Time    \t|\tUsername\t|\tHostname")
        for proc in processes:
            start_time = as_configured_timezone(proc.process_start_time, apply_time_format="%Y-%m-%d %H:%M:%S%z")
            username = proc.process_username[0]
            print("  {}\t    {}\t\t{}".format(start_time, username, proc.device_name))
        print()
    return



"""Universal Binary Store functions.
"""

import os
import io
import json
import time
import logging
import requests

from datetime import datetime
from typing import Dict, List, Union

from cbapi.connection import BaseAPI
from cbapi.errors import ServerError, ClientError, ObjectNotFoundError

LOGGER = logging.getLogger("cbinterface.psc.ubs")


def request_file_downloads(cb: BaseAPI, sha256hashes: List, expiration_seconds: int = 900) -> Dict:
    """Request file download URLs.

    If the UBS has a file for the sha256, it should return a URL to GET the file content.

    Args:
        cb: Carbon Black Cloud API connection with correct permissions.
        sha256hashes: list of sha256 hashes
        expiration_seconds: How many seconds the file content should be accessible.

    Returns:
        A dict object describing results.
    """
    url = f"/ubs/v1/orgs/{cb.credentials.org_key}/file/_download"
    request_data = {"sha256": sha256hashes, "expiration_seconds": expiration_seconds}
    try:
        result = cb.post_object(url, request_data)
        return result.json()
    except ServerError as e:
        LOGGER.error(f"Caught ServerError: {e}")
        return False
    except ClientError as e:
        LOGGER.warning(f"got ClientError: {e}")
        return False
    except ValueError:
        LOGGER.warning(f"got unexpected {result}")
        return False


def get_file_content(cb: BaseAPI, file_found_object: Dict, compressed=True):
    """Return the file content as a byte string.

    Args:
        cb: Carbon Black Cloud API object
        file_found_object: Details about the files that were found to
          be available for download

    Returns:
        The raw or compressed file content bytes.
    """
    import io
    import zipfile

    sha256 = file_found_object["sha256"]
    url = file_found_object["url"]
    LOGGER.debug(f"getting file content for sha256={sha256} from: {url}")
    compressed_content = b""
    result = requests.get(url, stream=True, proxies=cb.session.proxies)
    if result.status_code == 200:
        for chunk in result.iter_content(io.DEFAULT_BUFFER_SIZE):
            compressed_content += chunk
    else:
        LOGGER.error(f"got {result.status_code} attempting file get: {result}")
        return False

    if compressed:
        return compressed_content

    with zipfile.ZipFile(io.BytesIO(compressed_content)) as zp:
        if "filedata" not in zp.namelist():
            LOGGER.error(f"unexpected UBS compressed file content: missing 'filedata'")
            return False
        with zp.open("filedata") as fp:
            return fp.read()


def get_file(cb: BaseAPI, file_found_object: Dict, write_path=None, compressed=True):
    """Get file and write content to disk.

    Args:
        cb: Carbon Black Cloud API object
        file_found_object: Details about the files that were found to
          be available for download
        write_path: Where to write the file. Default is sha256 or sha256.zip
        compressed: If true, write as zip file.

    Returns:
        The write_path if a file was written to disk.
    """
    sha256 = file_found_object["sha256"]
    if write_path is None:
        if compressed:
            write_path = f"{sha256}.zip"
        else:
            write_path = sha256
    LOGGER.debug(f"writing file with sha256={sha256} to: {write_path}")
    with open(write_path, "wb") as fp:
        fp.write(get_file_content(cb, file_found_object, compressed=compressed))

    if os.path.exists(write_path):
        LOGGER.info(f" + Wrote: {write_path}")
        return write_path
    return None


def request_and_get_file(
    cb: BaseAPI, sha256_hash, expiration_seconds: int = 900, write_path=None, compressed=True, return_bytes=False
):
    """Request and download any content for file with sha256 hash.

    Args:
        cb: Carbon Black Cloud API object
        sha256_hash: A sha256 hash to request.
        expiration_seconds: How many seconds the file content should be accessible.
        compressed: If true, return zip compressed bytes; else, return bytes.

    Returns:
        The file content as a byte string, or the file path the content was written to.
    """
    file_request_results = request_file_downloads(cb, [sha256_hash])
    for error in file_request_results["error"]:
        LOGGER.warning(f"UBS had an 'intermittent' error and you should re-try for sha256: {error}")
    for not_found in file_request_results["not_found"]:
        LOGGER.info(f"UBS did not find result for sha256: {not_found}")
    file_found_object = file_request_results["found"][0]
    if return_bytes:
        if write_path:
            LOGGER.warning(f"nonsensical for write_path and return_bytes to both be set.")
        return get_file_content(cb, file_found_object, compressed=compressed)
    return get_file(cb, file_found_object, write_path=write_path, compressed=compressed)


def request_and_get_files(cb: BaseAPI, sha256hashes: List, expiration_seconds: int = 900, compressed=True):
    """Request and download zip compressed files found by the sha256 list.

    Args:
        cb: Carbon Black Cloud API object
        sha256hashes: A list of sha256 hashes to request.
        expiration_seconds: How many seconds the file content should be accessible.
        compressed: If true, write as zip files.

    Returns:
        list of file paths that were written.
    """
    written_file_paths = []
    file_request_results = request_file_downloads(cb, sha256hashes)
    for error in file_request_results["error"]:
        LOGGER.warning(f"UBS had an 'intermittent' error and you should re-try for sha256: {error}")
    for not_found in file_request_results["not_found"]:
        LOGGER.info(f"UBS did not find result for sha256: {not_found}")
    for file_found_object in file_request_results["found"]:
        fpath = get_file(cb, file_found_object, compressed=compressed)
        written_file_paths.append(fpath)

    return written_file_paths


def yield_file_metadata(cb: BaseAPI, sha256hashes: List):
    """Yield metadata available for matching files."""
    url = f"/ubs/v1/orgs/{cb.credentials.org_key}/sha256"
    for sha256 in sha256hashes:
        try:
            yield cb.get_object(f"{url}/{sha256}/metadata")
        except ObjectNotFoundError as e:
            err_message = json.loads(e.message)
            LOGGER.info(f"UBS: {err_message['error_code']}: {err_message['message']}")
        except ServerError as e:
            LOGGER.error(f"Caught ServerError: {e}")
        except ClientError as e:
            LOGGER.warning(f"got ClientError: {e}")
            return False
        except Exception as e:
            LOGGER.error(f"UNHANDLED: {e}")


def get_file_metadata(cb: BaseAPI, sha256hashes: List):
    """Return any metadata available for any matching files."""
    return list(yield_file_metadata(cb, sha256hashes))


def yield_device_summary(cb: BaseAPI, sha256hashes: List):
    """Yield an overview of the devices that executed the file."""
    url = f"/ubs/v1/orgs/{cb.credentials.org_key}/sha256"
    for sha256 in sha256hashes:
        try:
            yield cb.get_object(f"{url}/{sha256}/summary/device")
        except ObjectNotFoundError as e:
            err_message = json.loads(e.message)
            LOGGER.info(f"UBS: {err_message['error_code']}: {err_message['message']}")
        except ServerError as e:
            LOGGER.error(f"Caught ServerError: {e}")
        except ClientError as e:
            LOGGER.warning(f"got ClientError: {e}")
            return False
        except Exception as e:
            LOGGER.error(f"UNHANDLED: {e}")


def get_device_summary(cb: BaseAPI, sha256hashes: List):
    """Return an overview of the devices that executed the file."""
    return list(yield_device_summary(cb, sha256hashes))


def yield_signature_summary(cb: BaseAPI, sha256hashes: List, rows: int = 5):
    """Yield an overview of digital signature for a given SHA-256 hash

    This API will return a summary of the observed digital signature results
    for a given SHA-256 hash. The digital signature information for a binary
    may vary from one machine to another based on a variety of factors,
    including the presence of an up-to-date signature catalog on the host,
    system clock variations, ability to reach OCSP servers, custom root trust
    anchors, and more. Therefore, the results are ordered by prevalence, such
    that the most observed signatures will be returned first. The number of
    results are configurable, up to a max of 100 entries. Digital signatures
    can be recorded in a separate file (known as a “catalog” file) or embedded
    inside of the binary itself (an “embedded” signature). This signature
    API will capture the results of the endpoint’s verification of the digital
    signature associated with a given SHA-256 hash, stating whether that signature
    validation was based on a catalog file or an embedded signature. It is not
    uncommon for a catalog signature to be frequently updated, such that this API
    can return a wide variety of sign_timestamp values for a given SHA-256 hash.
    """
    url = f"/ubs/v1/orgs/{cb.credentials.org_key}/sha256"
    for sha256 in sha256hashes:
        try:
            yield cb.get_object(f"{url}/{sha256}/summary/signature?rows={rows}")
        except ObjectNotFoundError as e:
            err_message = json.loads(e.message)
            LOGGER.info(f"UBS: {err_message['error_code']}: {err_message['message']}")
        except ServerError as e:
            LOGGER.error(f"Caught ServerError: {e}")
        except ClientError as e:
            LOGGER.warning(f"got ClientError: {e}")
            return False
        except Exception as e:
            LOGGER.error(f"UNHANDLED: {e}")


def get_signature_summary(cb: BaseAPI, sha256hashes: List):
    """Return an overview of digital signature for a given SHA-256 hash."""
    return list(yield_signature_summary(cb, sha256hashes))


def yield_file_path_summary(cb: BaseAPI, sha256hashes: List, rows: int = 5):
    """Summary of the observed file paths for given SHA-256 hashes.

    This API will return a summary of the observed file paths. The results are
    ordered by prevalence, such that the most observed file path will be returned
    first. The number of results are configurable, up to a max of 100 entries.
    """
    url = f"/ubs/v1/orgs/{cb.credentials.org_key}/sha256"
    for sha256 in sha256hashes:
        try:
            yield cb.get_object(f"{url}/{sha256}/summary/file_path?rows={rows}")
        except ObjectNotFoundError as e:
            err_message = json.loads(e.message)
            LOGGER.info(f"UBS: {err_message['error_code']}: {err_message['message']}")
        except ServerError as e:
            LOGGER.error(f"Caught ServerError: {e}")
        except ClientError as e:
            LOGGER.warning(f"got ClientError: {e}")
            return False
        except Exception as e:
            LOGGER.error(f"UNHANDLED: {e}")


def get_file_path_summary(cb: BaseAPI, sha256hashes: List):
    """Summary of the observed file paths for given SHA-256 hashes."""
    return list(yield_file_path_summary(cb, sha256hashes))


def consolidate_metadata_and_summaries(cb: BaseAPI, sha256hashes: List):
    """Combine file metadata and all summary information for given SHA-256 hashes.

    Args:
      cb: A Cb API object
      sha256hashes: list of hashes to query about.
    Returns:
      A list of dictionaries, where each dictionary is a combined result
      for the respective sha256 hash.
    """
    results = []
    for sha256 in sha256hashes:
        sha256_data = {
            "sha256": sha256,
            "metadata": [],
            "device_sumamry": [],
            "signature_summary": [],
            "file_path_summary": [],
        }
        sha256_data["metadata"] = get_file_metadata(cb, [sha256])
        if sha256_data["metadata"]:
            sha256_data["device_summary"] = get_device_summary(cb, [sha256])
            sha256_data["signature_summary"] = get_signature_summary(cb, [sha256])
            sha256_data["file_path_summary"] = get_file_path_summary(cb, [sha256])
        results.append(sha256_data)
    return results



"""PSC Threathunter CLI functions."""

import os
import sys
import argparse
import datetime
import logging
import json
import yaml

from dateutil import tz

from typing import List, Union

from cbapi import __file__ as cbapi_file_path
from cbapi.errors import ObjectNotFoundError, MoreThanOneResultError, ClientError
from cbapi.psc import Device
from cbapi.psc.devices_query import DeviceSearchQuery
from cbapi.psc.threathunter import CbThreatHunterAPI, Process, Watchlist, Report, Feed
from cbapi.psc.threathunter.query import Query

from cbinterface.helpers import is_psc_guid, clean_exit, input_with_timeout
from cbinterface.psc.query import make_process_query, print_facet_histogram, yield_events
from cbinterface.psc.ubs import (
    request_and_get_files,
    get_file_metadata,
    get_device_summary,
    get_signature_summary,
    get_file_path_summary,
    consolidate_metadata_and_summaries,
)
from cbinterface.psc.intel import (
    convert_response_watchlists_to_psc_edr_watchlists,
    get_all_watchlists,
    get_watchlist,
    get_report,
    delete_report,
    get_report_with_IOC_status,
    print_report,
    interactively_update_report_ioc_query,
    convert_response_watchlists_to_single_psc_edr_watchlist,
    get_all_feeds,
    get_feed,
    get_feed_report,
    get_alert,
    get_all_alerts,
    update_alert_state,
    interactively_update_alert_state,
    get_watchlists_like_name,
    search_feed_names,
    is_ioc_ignored,
    ignore_ioc,
    activate_ioc,
    create_new_report_and_append_to_watchlist,
    write_basic_report_template,
)
from cbinterface.psc.device import (
    make_device_query,
    device_info,
    time_since_checkin,
    find_device_by_hostname,
    is_device_online,
)
from cbinterface.psc.process import (
    select_process,
    print_process_info,
    print_ancestry,
    print_process_tree,
    print_modloads,
    print_filemods,
    inspect_process_tree,
    print_netconns,
    print_regmods,
    print_crossprocs,
    print_childprocs,
    print_scriptloads,
    process_to_dict,
    format_event_data,
)
from cbinterface.commands import (
    PutFile,
    ProcessListing,
    GetFile,
    ListRegKeyValues,
    RegKeyValue,
    ExecuteCommand,
    ListDirectory,
    WalkDirectory,
    LogicalDrives,
    DeleteFile,
    KillProcessByID,
    KillProcessByName,
    DeleteRegistryKeyValue,
    DeleteRegistryKey,
    SetRegKeyValue,
    CreateRegKey,
    GetSystemMemoryDump,
)
from cbinterface.psc.sessions import (
    CustomLiveResponseSessionManager,
    get_session_by_id,
    device_live_response_sessions_by_device_id,
    all_live_response_sessions,
    get_session_commands,
    get_command_result,
    get_file_content,
    close_session_by_id,
)
from cbinterface.psc.enumerations import logon_history
from cbinterface.config import get_playbook_map
from cbinterface.scripted_live_response import build_playbook_commands, build_remediation_commands

LOGGER = logging.getLogger("cbinterface.psc.cli")


def toggle_device_quarantine(
    cb: CbThreatHunterAPI, devices: Union[DeviceSearchQuery, List[Device]], quarantine: bool
) -> bool:
    """Toggle device quarantine state.

    Args:
        devices: DeviceSearchQuery
        quarantine: set quarantine if True, else set quarantine to off state.
    """
    if len(devices) > 0:
        if len(devices) > 10 and quarantine:
            LOGGER.error(
                f"For now, not going to quarnantine {len(devices)} devices as a safe gaurd "
                f"to prevent mass device impact... use the GUI if you must."
            )
            return False
        verbiage = "quarantine" if quarantine else "NOT quarantine"
        emotion = "👀" if quarantine else "👏"
        LOGGER.info(f"setting {verbiage} on {len(devices)} devices... {emotion}")

        device_ids = []
        for d in devices:
            if d.quarantined == quarantine:
                LOGGER.warning(f"device {d.id}:{d.name} is already set to {verbiage}.")
                continue
            if not is_device_online(d):
                LOGGER.info(f"device {d.id}:{d.name} hasn't checked in for: {time_since_checkin(d, refresh=False)}")
                LOGGER.warning(f"device {d.id}:{d.name} appears offline 💤")
                LOGGER.info(f"device {d.id}:{d.name} will change quarantine state when it comes online 👌")
            device_ids.append(d.id)
            cb.device_quarantine(device_ids, quarantine)
        return True


def add_psc_arguments_to_parser(subparsers: argparse.ArgumentParser) -> None:
    """Given an argument parser subparser, build a psc specific parser."""
    # device query (psc)
    parser_sensor = subparsers.add_parser("device", aliases=["d"], help="Execute a device query (PSC).")
    parser_sensor.add_argument("device_query", help="the device query you'd like to execute. 'FIELDS' for help.")
    parser_sensor.add_argument(
        "-nw",
        "--no-warnings",
        action="store_true",
        default=False,
        help="Don't warn before printing large query results",
    )
    parser_sensor.add_argument(
        "-ad",
        "--all-details",
        action="store_true",
        default=False,
        help="Print all available process info (all fields).",
    )
    parser_sensor.add_argument(
        "-q",
        "--quarantine",
        action="store_true",
        default=False,
        help="Quarantine the devices returned by the query.",
    )
    parser_sensor.add_argument(
        "-uq",
        "--un_quarantine",
        action="store_true",
        default=False,
        help="UN-Quarantine the devices returned by the query.",
    )

    # UBS parser
    parser_ubs = subparsers.add_parser(
        "ubs", help="Interface with the Universal Binary Store (UBS) to download files and/or get information."
    )
    parser_ubs.add_argument(
        "--sha256",
        dest="sha256hashes",
        action="append",
        default=[],
        help="The SHA-256 hash of a file you're interested in. Use multiple times to build list.",
    )
    parser_ubs.add_argument(
        "--from-stdin", action="store_true", help="Read SHA-256 hashes piped from stdin to work with."
    )
    parser_ubs.add_argument(
        "-g",
        "--get-file",
        dest="ubs_get_file",
        action="store_true",
        help="Attempt to download file content for the SHA-256 hashes supplied by `--sha256`",
    )
    parser_ubs.add_argument(
        "-ds",
        "--get-device-summary",
        dest="ubs_get_device_summary",
        action="store_true",
        help="Get an overview of the devices that executed the file.",
    )
    parser_ubs.add_argument(
        "-ss",
        "--get-signature-summary",
        dest="ubs_get_signature_summary",
        action="store_true",
        help="Summary of the observed digital signature results for a given SHA-256 hashes.",
    )
    parser_ubs.add_argument(
        "-fps",
        "--get-file-path-summary",
        dest="ubs_get_file_path_summary",
        action="store_true",
        help="Summary of the observed file paths for a given SHA-256 hashes.",
    )
    parser_ubs.add_argument(
        "-i",
        "--get-metadata",
        dest="ubs_get_metadata",
        action="store_true",
        help="Get file metadata for give SHA-256 hashes.",
    )
    parser_ubs.add_argument(
        "-ci",
        "--combined-info",
        dest="ubs_combined_info",
        action="store_true",
        help="Combine metadata and summaries per SHA-256",
    )

    # intel parser
    parser_intel = subparsers.add_parser("intel", help="Intel Feeds, Watchlists, Reports, & IOCs")
    parser_intel.add_argument("--json", action="store_true", help="Return results as JSON.")

    intel_subparsers = parser_intel.add_subparsers(dest="intel_command")

    # intel watchlists
    parser_intel_watchlists = intel_subparsers.add_parser("watchlists", help="Interface with PSC Watchlists.")
    parser_intel_watchlists.add_argument("-lw", "--list-watchlists", action="store_true", help="List all watchlists.")
    parser_intel_watchlists.add_argument("-w", "--get-watchlist", action="store", help="Get watchlist by ID.")
    parser_intel_watchlists.add_argument(
        "-wn", "--watchlist-name-search", action="store", help="Search for watchlists by name."
    )
    parser_intel_watchlists.add_argument(
        "-wr", "--get-watchlist-report", action="store", help="Get a watchlist report by report ID."
    )
    parser_intel_watchlists.add_argument(
        "-dr", "--delete-watchlist-report", action="store", help="Delete watchlist report by ID."
    )
    parser_intel_watchlists.add_argument(
        "-wt",
        "--write-basic-threat-report-template",
        action="store_true",
        help="Write a basic singel query IOC threat report template.",
    )
    parser_intel_watchlists.add_argument(
        "--update-ioc-query",
        action="store",
        help="Update a query IOC for the given report ID/IOC id. format: report_id/ioc_id",
    )
    parser_intel_watchlists.add_argument("--get-ioc-status", action="store", help="Get active/ignore status of an IOC.")
    parser_intel_watchlists.add_argument("--ignore-ioc", action="store", help="Ignore IOC.")
    parser_intel_watchlists.add_argument("--activate-ioc", action="store", help="Activate IOC.")

    # create new threat reports for watchlists
    parser_intel_watchlists_subparsers = parser_intel_watchlists.add_subparsers(dest="intel_watchlist_command")
    parser_intel_watchlist_creation = parser_intel_watchlists_subparsers.add_parser(
        "new", help="Create new Threat Report for a Watchlist."
    )
    parser_intel_watchlist_creation.add_argument(
        "report_path", action="store", help="Path to JSON representation of new Threat Report."
    )
    parser_intel_watchlist_creation.add_argument(
        "-w", "--watchlist-id", required=True, action="store", help="The ID of a watchlist to append to."
    )

    # intel feeds
    parser_intel_feeds = intel_subparsers.add_parser("feeds", help="Interface with PSC Feeds.")
    parser_intel_feeds.add_argument("-lf", "--list-feeds", action="store_true", help="List all Feeds, public included.")
    parser_intel_feeds.add_argument(
        "-f",
        "--get-feed",
        action="store",
        help="Get Feed by ID. WARNING: Can return a lot of data if using the `--json` arg.",
    )
    parser_intel_feeds.add_argument(
        "-s", "--search-for-feed", action="store", help="Search the Feeds for feed names containing this value."
    )
    parser_intel_feeds.add_argument(
        "-fr",
        "--get-feed-report",
        action="store",
        help="Get specific Report from specific Feed. format: feed_id/report_id",
    )

    # alert parser plopped in here under intel
    parser_intel_alerts = intel_subparsers.add_parser("alerts", help="Interface with PSC Alerts.")
    parser_intel_alerts.add_argument(
        "-a", "--alert-id", dest="alert_ids", default=[], action="append", help="List alert IDs to work with."
    )
    parser_intel_alerts.add_argument("-g", "--get-alert", action="store_true", help="Get Alert information.")
    parser_intel_alerts.add_argument("-d", "--dismiss-alert", action="store_true", help="Dismiss an Alerts.")
    parser_intel_alerts.add_argument("-o", "--open-alert", action="store_true", help="Set Alerts to Open (un-dismiss).")
    parser_intel_alerts.add_argument(
        "-u", "--interactively-update-alert", action="store_true", help="Update Alerts interactively."
    )
    parser_intel_alerts.add_argument(
        "-r",
        "--remediation-state",
        action="store",
        help="An Alert remediation state to use with any state change actions.",
    )
    parser_intel_alerts.add_argument(
        "-c", "--comment", action="store", help="An Alert comment to use with any state change actions."
    )
    parser_intel_alerts.add_argument(
        "--from-stdin", action="store_true", help="Read alert IDs from stdin to work with."
    )

    intel_alerts_subparsers = parser_intel_alerts.add_subparsers(dest="intel_alerts_command")
    # alert search parser
    parser_intel_alerts_search = intel_alerts_subparsers.add_parser(
        "search", help="Search Alerts with lucene syntax queries and/or value searches."
    )
    parser_intel_alerts_search.add_argument("alert_query", action="store", help="The Alert search query.")
    # TODO Add other arguments to allow for searching via start & end times, specifying rows, start, alert state
    parser_intel_alerts_search.add_argument(
        "-cr",
        "--create_time-range",
        action="store",
        help="Only return alerts created over the previous time range. format:integer_quantity,time_unit ; time_unit in [s,m,h,d,w,y]",
    )
    parser_intel_alerts_search.add_argument(
        "-m",
        "--max-alerts-result",
        action="store",
        default=100,
        type=int,
        help="Only return up to this many alerts. Default=100. HINT: set to '0' to return all results.",
    )
    parser_intel_alerts_search.add_argument(
        "-as",
        "--alert-states",
        action="append",
        choices=["DISMISSED", "OPEN"],
        help="Only return Alerts in these states.",
    )

    # cb response to psc migration parser
    parser_intel_migration = intel_subparsers.add_parser(
        "migrate", help="Utilities for migrating response watchlists to PSC EDR intel."
    )
    parser_intel_migration.add_argument(
        "response_watchlist_json_data_path",
        help="Path to response watchlist json file. (see cbinterface response_watchlist",
    )
    parser_intel_migration.add_argument(
        "--one-for-one",
        action="store_true",
        help="Create a PSC Watchlist for every CbR watchlist that passes validation.",
    )
    parser_intel_migration.add_argument(
        "--many-to-one",
        action="store_true",
        help="Create a single PSC Watchlist containing all CbR watchlist queries that pass validation.",
    )


def execute_threathunter_arguments(cb: CbThreatHunterAPI, args: argparse.Namespace) -> bool:
    """The logic to execute psc ThreatHunter specific command line arguments.

    Args:
        cb: CbThreatHunterAPI
        args: parsed argparse namespace
    Returns:
        True or None on success, False on failure.
    """
    if not isinstance(cb, CbThreatHunterAPI):
        LOGGER.critical(f"Requires Cb PSC based API. Got '{product}' API.")
        return False

    # UBS #
    if args.command == "ubs":
        if args.from_stdin:
            args.sha256hashes.extend([line.strip() for line in sys.stdin])

        if args.sha256hashes:

            set_ubs_args = [arg for arg, value in vars(args).items() if arg.startswith("ubs_") and value is True]
            if not set_ubs_args:
                LOGGER.debug(f"seting ubs metadata argument as default.")
                args.ubs_get_metadata = True

            if args.ubs_get_file:
                request_and_get_files(cb, sha256hashes=args.sha256hashes)
            if args.ubs_get_device_summary:
                summary = get_device_summary(cb, args.sha256hashes)
                if summary:
                    print(json.dumps(summary, indent=2))
            if args.ubs_get_signature_summary:
                summary = get_signature_summary(cb, args.sha256hashes)
                if summary:
                    print(json.dumps(summary, indent=2))
            if args.ubs_get_file_path_summary:
                summary = get_file_path_summary(cb, args.sha256hashes)
                if summary:
                    print(json.dumps(summary, indent=2))
            if args.ubs_get_metadata:
                # this is default if no arguments are specified with the sha256(s)
                file_metadata = get_file_metadata(cb, sha256hashes=args.sha256hashes)
                if file_metadata:
                    print(json.dumps(file_metadata, indent=2))
            if args.ubs_combined_info:
                results = consolidate_metadata_and_summaries(cb, args.sha256hashes)
                if results:
                    print(json.dumps(results, indent=2))
        else:
            LOGGER.error(f"You must specify at least one sha256 with the `--sha256` argument.")
            return False

        return True

    # Intel #
    if args.command == "intel":
        if args.intel_command == "alerts":

            if args.intel_alerts_command == "search":  #'device_name': ['XW7R17'],
                # NOTE TODO: implement start and end time argparse options
                # criteria = {'last_event_time': {'start': '2021-04-13T19:39:30.054855+00:00', 'end': '2021-04-14T19:39:30.054855+00:00'}, 'workflow': ['OPEN']}
                if args.create_time_range:
                    criteria["create_time"] = {"range": f"-{args.create_time_range}"}
                results = get_all_alerts(
                    cb, query=args.alert_query, workflow_state=args.alert_states, max_results=args.max_alerts_result
                )
                if results:
                    print(json.dumps(results, indent=2))

                return True

            if args.from_stdin:
                args.alert_ids.extend([line.strip().strip('"') for line in sys.stdin])

            if not args.alert_ids:
                LOGGER.error(f"You have to supply at least one alert ID.")
                return False

            if args.get_alert:
                alerts = [get_alert(cb, alert_id) for alert_id in args.alert_ids]
                if alerts:
                    print(json.dumps(alerts, indent=2))

            if args.open_alert:
                results = [
                    update_alert_state(
                        cb, alert_id, state="OPEN", remediation_state=args.remediation_state, comment=args.comment
                    )
                    for alert_id in args.alert_ids
                ]
                if result:
                    print(json.dumps(results, indent=2))

            if args.dismiss_alert:
                results = [
                    update_alert_state(
                        cb,
                        alert_id,
                        state="DISMISSED",
                        remediation_state=args.remediation_state,
                        comment=args.comment,
                    )
                    for alert_id in args.alert_ids
                ]
                if results:
                    print(json.dumps(results, indent=2))

            if args.interactively_update_alert:
                results = [interactively_update_alert_state(cb, alert_id) for alert_id in args.alert_ids]
                if results:
                    print(json.dumps(results, indent=2))

        if args.intel_command == "migrate":
            response_watchlists = None
            with open(args.response_watchlist_json_data_path, "r") as fp:
                response_watchlists = json.load(fp)

            if args.one_for_one:
                results = convert_response_watchlists_to_psc_edr_watchlists(cb, response_watchlists)
                LOGGER.info(
                    f"created {len(results)} PSC watchlists from {len(response_watchlists)} Response watchlists."
                )
                print("Created watchlists:")
                for wl in results:
                    print(f" + ID={wl['id']} - Title={wl['name']}")

            if args.many_to_one:
                watchlist = convert_response_watchlists_to_single_psc_edr_watchlist(cb, response_watchlists)
                if not watchlist:
                    return False
                LOGGER.info(
                    f"Created \"{watchlist['name']}\" containing {len(watchlist['report_ids'])} intel reports based on {len(response_watchlists)} Response watchlists."
                )

        if args.intel_command == "watchlists":
            if args.intel_watchlist_command == "new":
                report_data = {}
                if not os.path.exists(args.report_path):
                    LOGGER.error(f"{args.report_path} does not exist.")
                    return False
                with open(args.report_path, "r") as fp:
                    report_data = json.load(fp)
                if not report_data:
                    LOGGER.error(f"failed to load report data")
                    return False
                watchlist_data = create_new_report_and_append_to_watchlist(cb, args.watchlist_id, report_data)
                if watchlist_data:
                    LOGGER.info(f"successfully appended new threat report to watchlist.")
                return True

            if args.write_basic_threat_report_template:
                result = write_basic_report_template()
                if result:
                    LOGGER.info(f"wrote: {result}")
                return result

            if args.list_watchlists:
                watchlists = get_all_watchlists(cb)
                if args.json:
                    print(json.dumps(watchlists, indent=2))
                else:
                    for wl in watchlists:
                        print(Watchlist(cb, initial_data=wl))
                        print()

            if args.get_watchlist:
                watchlist = get_watchlist(cb, args.get_watchlist)
                if watchlist:
                    print(json.dumps(watchlist, indent=2))

            if args.watchlist_name_search:
                watchlists = get_watchlists_like_name(cb, args.watchlist_name_search)
                if watchlists:
                    if args.json:
                        print(json.dumps(watchlists, indent=2))
                    else:
                        for wl in watchlists:
                            print(Watchlist(cb, initial_data=wl))
                            print()

            if args.get_watchlist_report:
                if args.json:
                    print(json.dumps(get_report_with_IOC_status(cb, args.get_watchlist_report), indent=2))
                else:
                    report = get_report_with_IOC_status(cb, args.get_watchlist_report)
                    if report:
                        print_report(report)  # specifically helpful with query based IOCs

            if args.delete_watchlist_report:
                result = delete_report(cb, args.delete_watchlist_report)
                if result.status_code == 204:
                    LOGGER.info(f"deleted watchlist report")

            if args.update_ioc_query:
                report_id, ioc_id = args.update_ioc_query.split("/", 1)
                updated_report = interactively_update_report_ioc_query(cb, report_id, ioc_id)
                if updated_report:
                    LOGGER.info(f"Query IOC ID={ioc_id} of report ID={report_id} successfully updated.")

            if args.get_ioc_status:
                report_id, ioc_id = args.get_ioc_status.split("/", 1)
                status = is_ioc_ignored(cb, report_id, ioc_id, check_existence=True)
                status = "IGNORED" if status else "ACTIVE"
                print(f"IOC ID={ioc_id} in Report ID={report_id} is {status}")

            if args.ignore_ioc:
                report_id, ioc_id = args.ignore_ioc.split("/", 1)
                status = ignore_ioc(cb, report_id, ioc_id)
                status = "IGNORED" if status else "ACTIVE"
                print(f"IOC ID={ioc_id} in Report ID={report_id} is {status}")

            if args.activate_ioc:
                report_id, ioc_id = args.activate_ioc.split("/", 1)
                status = activate_ioc(cb, report_id, ioc_id)
                status = "ACTIVE" if status else "IGNORED"
                print(f"IOC ID={ioc_id} in Report ID={report_id} is {status}")

        if args.intel_command == "feeds":
            if args.list_feeds:
                feeds = get_all_feeds(cb)
                if not feeds:
                    return None
                if args.json:
                    print(json.dumps(feeds, indent=2))
                else:
                    for f in feeds:
                        print(Feed(cb, initial_data=f))
                        print()

            if args.get_feed:
                feed = get_feed(cb, args.get_feed)
                if not feed:
                    return None
                if args.json:
                    print(json.dumps(feed, indent=2))
                else:
                    print(Feed(cb, initial_data=feed))

            if args.search_for_feed:
                feeds = search_feed_names(cb, args.search_for_feed)
                if not feeds:
                    return None
                if args.json:
                    print(json.dumps(feeds, indent=2))
                else:
                    for f in feeds:
                        print(Feed(cb, initial_data=f))
                        print()

            if args.get_feed_report:
                try:
                    feed_id, report_id = args.get_feed_report.split("/", 1)
                except ValueError:
                    feed_id, report_id = args.get_feed_report.split("-", 1)
                report = get_feed_report(cb, feed_id, report_id)
                print(json.dumps(report, indent=2))

        return True

    # Device Quering #
    if args.command and args.command.startswith("d"):
        LOGGER.info(f"searching {args.environment} environment for device query: {args.device_query}...")
        if args.device_query.upper() == "FIELDS":
            device_meta_file = os.path.join(os.path.dirname(cbapi_file_path), "psc/defense/models/deviceInfo.yaml")
            model_data = {}
            with open(device_meta_file, "r") as fp:
                model_data = yaml.safe_load(fp.read())
            possibly_searchable_props = list(model_data["properties"].keys())
            print("Device model fields:")
            for field_name in list(model_data["properties"].keys()):
                print(f"\t{field_name}")
            return True

        if args.quarantine and args.un_quarantine:
            LOGGER.error("quarantine AND un-quarantine? 🤨 Won't do it.")
            return False

        devices = make_device_query(cb, args.device_query)
        if not devices:
            return None

        # Quarantine?
        if args.quarantine:
            toggle_device_quarantine(cb, devices, True)
        elif args.un_quarantine:
            toggle_device_quarantine(cb, devices, False)

        # don't display large results by default
        print_results = True
        if not args.no_warnings and len(devices) > 10:
            prompt = "Print all results? (y/n) [y] "
            print_results = input_with_timeout(prompt, default="y")
            print_results = True if print_results.lower() == "y" else False

        if len(devices) > 0 and print_results:
            print("\n------------------------- PSC DEVICE RESULTS -------------------------")
            for device in devices:
                if args.all_details:
                    print()
                    print(device)
                else:
                    print(device_info(device))
            print()
        return True

    # Process Quering #
    if args.command and (args.command.startswith("q") or args.command == "pq"):
        LOGGER.info(f"searching {args.environment} environment..")

        # format datetimes as needed
        format_string = "%Y-%m-%d %H:%M:%S"
        if args.start_time and "T" in args.start_time or args.last_time and "T" in args.last_time:
            format_string = "%Y-%m-%dT%H:%M:%S"
        args.start_time = (
            datetime.datetime.strptime(args.start_time, format_string) if args.start_time else args.start_time
        )
        args.last_time = datetime.datetime.strptime(args.last_time, format_string) if args.last_time else args.last_time
        processes = make_process_query(
            cb,
            args.query,
            start_time=args.start_time,
            last_time=args.last_time,
            raise_exceptions=False,
            validate_query=True,
        )

        if args.facets:
            LOGGER.info("getting facet data...")
            print_facet_histogram(processes)
            # NOTE TODO - pick this v2 back up and see if it's more efficient to use
            # knowing we have to remember the childproc_name facet data we like.
            # from cbinterface.psc.query import print_facet_histogram_v2
            # print_facet_histogram_v2(cb, args.query)

        # don't display large results by default
        print_results = True
        if not args.no_warnings and len(processes) > 10:
            prompt = "Print all results? (y/n) [y] "
            print_results = input_with_timeout(prompt, default="y")
            print_results = True if print_results.lower() == "y" else False

        if len(processes) > 0 and print_results:
            print("\n------------------------- QUERY RESULTS -------------------------")
            for proc in processes:
                print("  -------------------------")
                if args.all_details:
                    print(proc)
                else:
                    print_process_info(proc, raw_print=args.all_details, header=False)

        return True

    # Enumerations #
    if args.command and args.command == "enumerate":
        if args.logon_history:
            logon_history(cb, args.logon_history)
            return

    # Process Inspection #
    if args.command and (args.command == "proc" or args.command.startswith("i")):
        process_id = args.process_guid_options
        if not is_psc_guid(process_id):
            # check to see if the analyst passed a local file path, which we assume is a local process json file
            # if os.path.exists(args.process_guid_options):
            # XXX NOTE: create functionality sourced from process json file?
            LOGGER.error(f"{process_id} is not in the form of a CbThreathunter process guid.")
            return False

        try:
            # proc = Process(cb, process_id)
            proc = select_process(cb, process_id)
            if not proc:
                LOGGER.warning(f"Process data does not exist for GUID={process_id}")
                return False
        except Exception as e:
            LOGGER.error(f"unexpected problem finding process: {e}")
            return False

        # format datetimes as needed
        format_string = "%Y-%m-%d %H:%M:%S"
        if args.start_time and "T" in args.start_time or args.end_time and "T" in args.end_time:
            format_string = "%Y-%m-%dT%H:%M:%S"
        args.start_time = (
            datetime.datetime.strptime(args.start_time, format_string).replace(tzinfo=tz.gettz("GMT"))
            if args.start_time
            else args.start_time
        )
        args.end_time = (
            datetime.datetime.strptime(args.end_time, format_string).replace(tzinfo=tz.gettz("GMT"))
            if args.end_time
            else args.end_time
        )

        if args.event_search:
            for event in yield_events(
                proc, query=args.event_search, start_time=args.start_time, end_time=args.end_time
            ):
                if args.raw_print_events:
                    print(json.dumps(event, default=str, indent=2, sort_keys=True))
                else:
                    print(format_event_data(event))
            return True

        if args.json:
            print(
                json.dumps(
                    process_to_dict(proc, start_time=args.start_time, end_time=args.end_time, event_rows=2000),
                    default=str,
                )
            )
            return

        all_inspection_args = [iarg for iarg in vars(args).keys() if iarg.startswith("inspect_")]
        set_inspection_args = [
            iarg for iarg, value in vars(args).items() if iarg.startswith("inspect_") and value is True
        ]
        if not set_inspection_args:
            LOGGER.debug(f"seting all inspection arguments.")
            for iarg in all_inspection_args:
                args.__setattr__(iarg, True)

        if args.walk_and_inspect_tree:
            inspect_process_tree(
                proc,
                info=args.inspect_proc_info,
                filemods=args.inspect_filemods,
                netconns=args.inspect_netconns,
                regmods=args.inspect_regmods,
                modloads=args.inspect_modloads,
                crossprocs=args.inspect_crossprocs,
                children=args.inspect_children,
                scriptloads=args.inspect_scriptloads,
                raw_print=args.raw_print_events,
                start_time=args.start_time,
                end_time=args.end_time,
            )
            return True

        if args.inspect_process_ancestry:
            print_ancestry(proc)
            print()
        if args.inspect_process_tree:
            print_process_tree(proc, start_time=args.start_time, end_time=args.end_time)
            print()
        if args.inspect_proc_info:
            print_process_info(proc, raw_print=args.raw_print_events)
        if args.inspect_filemods:
            print_filemods(proc, raw_print=args.raw_print_events, start_time=args.start_time, end_time=args.end_time)
        if args.inspect_netconns:
            print_netconns(proc, raw_print=args.raw_print_events, start_time=args.start_time, end_time=args.end_time)
        if args.inspect_regmods:
            print_regmods(proc, raw_print=args.raw_print_events, start_time=args.start_time, end_time=args.end_time)
        if args.inspect_modloads:
            print_modloads(proc, raw_print=args.raw_print_events, start_time=args.start_time, end_time=args.end_time)
        if args.inspect_crossprocs:
            print_crossprocs(proc, raw_print=args.raw_print_events, start_time=args.start_time, end_time=args.end_time)
        if args.inspect_children:
            print_childprocs(proc, raw_print=args.raw_print_events, start_time=args.start_time, end_time=args.end_time)
        if args.inspect_scriptloads:
            print_scriptloads(proc, raw_print=args.raw_print_events, start_time=args.start_time, end_time=args.end_time)

        return True

    # Live Response Actions #
    if args.command and (args.command.lower() == "lr" or args.command.lower().startswith("live")):
        # create a LR session manager
        session_manager = CustomLiveResponseSessionManager(cb, custom_session_keepalive=True)
        # store a list of commands to execute on this device
        commands = []

        LOGGER.info(f"searching for device...")
        device = None
        try:  # if device.id
            device = Device(cb, args.name_or_id)
        except ClientError:
            device = find_device_by_hostname(cb, args.name_or_id)

        if not device:
            LOGGER.info(f"could not find a device.")
            return None

        if args.execute_command:
            # XXX expand this for more flexibiliy by making an execute parser
            # that can accept more arugments to pass to ExecuteCommand
            cmd = ExecuteCommand(args.execute_command)
            commands.append(cmd)
            LOGGER.info(f"recorded command: {cmd}")

        # Quarantine?
        if args.quarantine:
            if toggle_device_quarantine(cb, [device], True):
                LOGGER.info(f"Device {device.id}:{device.name} is set to quarantine.")
        elif args.un_quarantine:
            if toggle_device_quarantine(cb, [device], False):
                LOGGER.info(f"Device {device.id}:{device.name} is set to NOT quarantine.")

        # Put File #
        if args.live_response_command and args.live_response_command.lower() == "put":
            cmd = PutFile(args.local_filepath, args.sensor_write_filepath)
            commands.append(cmd)
            LOGGER.info(f"recorded command: {cmd}")

        if args.create_regkey:
            cmd = CreateRegKey(args.create_regkey)
            commands.append(cmd)
            LOGGER.info(f"recorded command: {cmd}")
            if args.set_regkey_value:
                cmd = SetRegKeyValue(args.create_regkey, args.set_regkey_value)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

        # Sensor Collection #
        if args.live_response_command and args.live_response_command.lower() == "collect":
            if args.sensor_info:
                print(sensor_info(sensor))

            if args.process_list:
                cmd = ProcessListing()
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.list_directory:
                cmd = ListDirectory(args.list_directory)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.walk_directory:
                cmd = WalkDirectory(args.walk_directory)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.file:
                cmd = GetFile(args.file)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.regkeypath:
                cmd = ListRegKeyValues(args.regkeypath)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.regkeyvalue:
                cmd = RegKeyValue(args.regkeyvalue)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.drives:
                cmd = LogicalDrives()
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.memdump:
                cmd = GetSystemMemoryDump()
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

        # Sensor Remediation #
        if args.live_response_command and args.live_response_command == "remediate":
            if args.delete_file_path:
                cmd = DeleteFile(args.delete_file_path)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.kill_process_name:
                cmd = KillProcessByName(args.kill_process_name)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.delete_regkeyvalue:
                cmd = DeleteRegistryKeyValue(args.delete_regkeyvalue)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.delete_entire_regkey:
                cmd = DeleteRegistryKey(args.delete_entire_regkey)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.remediation_script:
                remediation_commands = build_remediation_commands(args.remediation_script)
                LOGGER.info(f"created {len(remediation_commands)} remediation commands from {args.remediation_script}")
                commands.extend(remediation_commands)

        # Playbook execution #
        if args.live_response_command and (
            args.live_response_command.startswith("play") or args.live_response_command == "pb"
        ):
            if args.playbook_configpath:
                playbook_commands = build_playbook_commands(args.playbook_configpath)
                commands.extend(playbook_commands)
                LOGGER.info(f"loaded {len(playbook_commands)} playbook commands.")
            if args.playbook_name:
                playbook_data = get_playbook_map()[args.playbook_name]
                playbook_path = playbook_data["path"]
                playbook_commands = build_playbook_commands(playbook_path)
                commands.extend(playbook_commands)
                LOGGER.info(f"loaded {len(playbook_commands)} playbook commands.")

        # Handle LR commands #
        if commands:
            timeout = 1200  # default 20 minutes (same used by Cb)
            if not is_device_online(device):
                # Decision point: if the device is NOT online, give the analyst and option to wait
                LOGGER.warning(f"{device.id}:{device.name} is offline.")
                prompt = "Would you like to wait for the host to come online? (y/n) [y] "
                wait = input_with_timeout(prompt, default="y")
                wait = True if wait.lower() == "y" else False
                if not wait:
                    return None
                prompt = "How many days do you want to wait? [Default is 7 days] "
                timeout = input_with_timeout(prompt, default=7)
                if isinstance(timeout, str):
                    timeout = int(timeout)
                if timeout > 30:
                    LOGGER.warning(f"{timeout} days is a long time. Restricting to max of 30 days.")
                    timeout = 30

                # 86400 seconds in a day
                timeout = timeout * 86400

            if not session_manager.wait_for_active_session(device, timeout=timeout):
                LOGGER.error(f"reached timeout waiting for active session.")
                return False

            # we have an active session, issue the commands.
            for command in commands:
                session_manager.submit_command(command, device)

        if session_manager.commands:
            # Wait for issued commands to complete and process any results.
            session_manager.process_completed_commands()

    # Direct Session Interaction #
    if args.command and args.command.startswith("sess"):
        cblr = CbThreatHunterAPI(url=cb.credentials.url, token=cb.credentials.lr_token, org_key=cb.credentials.org_key)

        # if args.list_all_sessions:
        # Not implemented with PSC
        # if args.list_sensor_sessions:
        # Not implemented with PSC

        if args.get_session_command_list:
            print(json.dumps(get_session_commands(cblr, args.get_session_command_list), indent=2, sort_keys=True))

        if args.get_session:
            print(json.dumps(get_session_by_id(cblr, args.get_session), indent=2, sort_keys=True))

        if args.close_session:
            print(json.dumps(close_session_by_id(cblr, args.close_session), indent=2, sort_keys=True))

        if args.get_command_result:
            session_id, device_id, command_id = args.get_command_result.split(":", 2)
            session_id = f"{session_id}:{device_id}"
            print(json.dumps(get_command_result(cblr, session_id, command_id), indent=2, sort_keys=True))

        if args.get_file_content:
            session_id, device_id, file_id = args.get_file_content.split(":", 2)
            session_id = f"{session_id}:{device_id}"
            get_file_content(cblr, session_id, file_id)



"""Everthing ThreatHunter process related.
"""

import datetime
import logging

from io import StringIO
from contextlib import redirect_stdout
from typing import Dict, Union, List

from cbinterface.psc.query import yield_events

from cbapi.psc.threathunter import CbThreatHunterAPI, Process, Event
from cbapi.errors import ObjectNotFoundError

from cbinterface.helpers import as_configured_timezone, get_os_independent_filepath

LOGGER = logging.getLogger("cbinterface.psc.process")

ALL_EVENT_TYPES = [
    "filemod",
    "netconn",
    "netconn_proxy" "regmod",
    "modload",
    "crossproc",
    "childproc",
    "scriptload",
    "fileless_scriptload",
]


def load_process(p: Process) -> Process:
    """Load any process meta-data that exists or return None."""
    try:
        return Process.new_object(p._cb, p.summary._info["process"])
    except RecursionError:
        LOGGER.warning(f"RecursionError occurred loading process details.. loading incomplete details.")
        url = f"/api/investigate/v1/orgs/{p._cb.credentials.org_key}/processes/summary"
        summary = p._cb.get_object(url, query_parameters={"process_guid": p.process_guid})
        return Process.new_object(p._cb, summary["process"])
    except ObjectNotFoundError:
        LOGGER.debug(f"Process data does not exist for GUID={p.get('process_guid')}")
        return None


def select_process(cb: CbThreatHunterAPI, process_guid: str):
    """Select and load a Cb processes with a single API call.

    MUCH faster than using the suggested method.

    Args:
        cb: a CbThreatHunterAPI object
        process_guid: a Cb PSC GUID in the form of a string.
    Returns:
        A threathunter.Process or None
    """
    # Create empty Process. `force_init` is not supported (underlying PSC model is unrefreshable)
    # Use the process object (see load_process) to load a new process object containing the completed process meta-data.
    return load_process(Process(cb, process_guid))


def is_process_loaded(p: Process) -> bool:
    """Return True if the process has data."""
    if len(p._info.keys()) < 3:
        return False
    return True


def process_to_dict(
    p: Process,
    max_events: int = None,
    start_time: datetime.datetime = None,
    end_time: datetime.datetime = None,
    event_rows=2000,
) -> Dict:
    """Convert process and it's events to a dictionary.

    Args:
      p: Carbon Black Cloud Process
      max_events: a limit on the amount of process events to add.
    Returns:
      Dictionary representation of the process
    """
    # TODO: add start & end args so process events during a specific
    #  time window can be pulled. This could be added for several functions on this page.
    process = {}
    if not is_process_loaded(p):
        p = load_process(p)
    process["info"] = p._info
    process["events"] = {}
    event_count = 0
    for event in yield_events(p, start_time=start_time, end_time=end_time, rows=event_rows):
        if event.get("event_type") not in process["events"]:
            process["events"][event["event_type"]] = []
        process["events"][event["event_type"]].append(event)
        event_count += 1
        if max_events is not None:
            assert isinstance(max_events, int)
            if event_count >= max_events:
                LOGGER.info(f"max event limit of {max_events} reached.")
                break

    process["process_ancestry"] = StringIO()
    with redirect_stdout(process["process_ancestry"]):
        print_ancestry(p)
    process["process_ancestry"] = process["process_ancestry"].getvalue()

    process["process_tree"] = StringIO()
    with redirect_stdout(process["process_tree"]):
        print_process_tree(p, start_time=start_time, end_time=end_time)
    process["process_tree"] = process["process_tree"].getvalue()

    return process


def print_process_info(proc: Process, yield_strings: bool = False, raw_print=False, header=True):
    """Analyst friendly custom process data format.

    Args:
        proc: CbTH Process (fully initialized)
        yield_strings: return string if True, else print it to stdout.
    Returns: string or None
    """

    if not is_process_loaded(proc):
        proc = load_process(proc)

    txt = ""
    if header:
        txt += "------ INFO ------\n"
    if raw_print:
        txt = str(proc)
    else:
        txt += f"  Process GUID: {proc.get('process_guid')}\n"
        process_name = get_os_independent_filepath(proc.get("process_name", "None")).name
        txt += f"  Process Name: {process_name}\n"
        process_pid = [str(_) for _ in proc.get("process_pid", [])]
        txt += f"  Process PID: {', '.join(process_pid)}\n"
        txt += f"  Process MD5: {proc.get('process_md5')}\n"
        txt += f"  Process SHA256: {proc.get('process_sha256')}\n"
        txt += f"  Process Path: {proc.get('process_name')}\n"
        txt += f"  Process Terminated: {proc.get('process_terminated')}\n"
        txt += f"  Start Time: {as_configured_timezone(proc.get('process_start_time', ''))}\n"
        process_command_line = (
            proc.process_cmdline[0]
            if proc.get("process_cmdline") and len(proc.process_cmdline) == 1
            else proc.get("process_cmdline")
        )
        txt += f"  Command Line: {process_command_line}\n"
        txt += f"  Process Reputation: {proc.get('process_reputation')}\n"
        txt += f"  Parent Name: {proc.get('parent_name')}\n"
        txt += f"  Parent GUID: {proc.get('parent_guid')}\n"
        parent_sha256 = next((hsh for hsh in proc.get("parent_hash", []) if len(hsh) == 64), None)
        txt += f"  Parent SHA256: {parent_sha256}\n"
        txt += f"  Process Username: {proc.get('process_username')}\n"
        txt += f"  Device Username: {proc.get('device_username')}\n"
        txt += f"  Device ID: {proc.get('device_id')}\n"
        txt += f"  Device Name: {proc.get('device_name')}\n"
        txt += f"  Device OS: {proc.get('device_os')}\n"
        txt += f"  External IP: {proc.get('device_external_ip')}\n"
        txt += f"  Internal IP: {proc.get('device_internal_ip')}\n"
    if yield_strings:
        return txt
    txt += "\n"
    print(txt)


def print_ancestry(p: Process, max_depth=0, depth=0):
    """Print the process ancestry for this process."""

    if max_depth and depth > max_depth:
        return

    if depth == 0:
        print("\n------ Process Ancestry ------")
        print()

    if not is_process_loaded(p):
        p = load_process(p)

    start_time = as_configured_timezone(p.get("process_start_time", ""))
    command_line = (
        p.process_cmdline[0] if p.get("process_cmdline") and len(p.process_cmdline) == 1 else p.get("process_cmdline")
    )
    print(f"{'  '*(depth + 1)}{start_time}: {command_line} | {p.process_guid}")

    if p.get("parent_guid"):
        parent = select_process(p._cb, p.parent_guid)
        if parent:
            print_ancestry(parent, max_depth=max_depth, depth=depth + 1)


def print_process_tree(
    p: Union[Process, Dict],
    max_depth=0,
    depth=0,
    start_time: datetime.datetime = None,
    end_time: datetime.datetime = None,
):
    """Print the process tree."""

    if max_depth and depth > max_depth:
        return

    if depth == 0:
        if isinstance(p, Process) and not is_process_loaded(p):
            p = load_process(p)

        print("\n------ Process Execution Tree ------")
        print()

        command_line = (
            p.process_cmdline[0]
            if p.get("process_cmdline") and len(p.get("process_cmdline")) == 1
            else p.get("process_cmdline")
        )
        print(f"  {'  '*(depth+1)}{command_line}  | {p.get('process_guid')}")
    else:
        process_name = get_os_independent_filepath(p.get("childproc_name", "None")).name
        command_line = p.get("childproc_cmdline") if p.get("childproc_cmdline") else p.get("childproc_name")
        print(f"  {'  '*(depth+1)}{process_name}:  {command_line}  | {p.get('childproc_process_guid')}")

    try:
        for child in yield_events(p, criteria={"event_type": ["childproc"]}, start_time=start_time, end_time=end_time):
            child["_cb"] = p.get("_cb")
            child["process_guid"] = child.get("childproc_process_guid")
            print_process_tree(child, max_depth=max_depth, depth=depth + 1)
    except RecursionError:
        LOGGER.warning(f"hit RecursionError walking process tree.. stopping here")
        print(" [!] reached recursion limit walking process tree ...")
    except ObjectNotFoundError:
        LOGGER.warning(f"got 404 object not found for child process")


def get_events_by_type(
    p: Union[Process, Dict],
    event_types: List[str],
    return_dict=False,
    start_time: datetime.datetime = None,
    end_time: datetime.datetime = None,
):
    """Return process events by event type.

    One of filemod, netconn, regmod, modload, crossproc, childproc, scriptload,
        fileless_scriptload, netconn_proxy
    """

    if isinstance(p, dict):
        # processing from json file.
        # NOTE time frame is not enforced here as we assume that happend when the json was written.
        assert "events" in p
        for event_type in event_types:
            for event in p["events"].get(event_type, []):
                yield event
        return

    try:
        # NOTE: when there are thousands of events... this under-the-hood will get ALL of them before returning..
        # return p.events(event_type=event_type)
        # So using our own code to yield events:
        for event in yield_events(p, criteria={"event_type": event_types}, start_time=start_time, end_time=end_time):
            if return_dict:
                yield event
            else:
                yield Event(p._cb, initial_data=event)
    except Exception as e:
        LOGGER.error(f"failed to get events: {e}")
        return


def format_filemod(fm: Union[Event, Dict]):
    """Format filemod event into single line."""
    _action_summary = [action[len("ACTION_") :] for action in fm.get("filemod_action", [])]
    _edge_actions = [action for action in _action_summary if not action.startswith("FILE")]
    _fm_action_summary = [action[len("FILE_") :] for action in _action_summary if action.startswith("FILE")]
    _fm_action_summary.extend(_edge_actions)
    action_summary = ",".join(_fm_action_summary)
    fm_sha256 = f" , sha256:{fm.get('filemod_sha256')}" if fm.get("filemod_sha256") else ""
    return (
        f" @{as_configured_timezone(fm.get('event_timestamp'))}: |{action_summary}| {fm.get('filemod_name')}{fm_sha256}"
    )


def print_filemods(
    p: Union[Process, Dict], raw_print=False, start_time: datetime.datetime = None, end_time: datetime.datetime = None
):
    """Print file modifications.

    one or more of ACTION_INVALID, ACTION_FILE_CREATE, ACTION_FILE_WRITE, ACTION_FILE_DELETE, ACTION_FILE_LAST_WRITE, ACTION_FILE_MOD_OPEN,
     ACTION_FILE_RENAME, ACTION_FILE_UNDELETE, ACTION_FILE_TRUNCATE, ACTION_FILE_OPEN_READ, ACTION_FILE_OPEN_WRITE, ACTION_FILE_OPEN_DELETE,
     ACTION_FILE_OPEN_EXECUTE, ACTION_FILE_READ
    """
    print("------ FILEMODS ------")
    for fm in get_events_by_type(p, ["filemod"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(fm)
            continue
        print(format_filemod(fm))
    print()


def format_netconn(nc: Union[Event, Dict]):
    """Format netconn or netconn_proxy event into a single line."""
    import ipaddress
    import socket, struct

    action = (
        nc.get("netconn_action", "")[len("ACTION_CONNECTION_") :]
        if nc.get("netconn_action", "").startswith("ACTION_CONNECTION_")
        else nc.get("netconn_action")
    )
    if action == "CREATE":
        # same behavior as PSC GUI
        action = "ESTABLISHED"
    action = action.capitalize()
    protocol = nc.get("netconn_protocol", "")
    if protocol.startswith("PROTO_"):
        protocol = protocol[len("PROTO_") :]
    direction = "inbound" if nc.get("netconn_inbound") else "outbound"

    # NOTE: Cb stores ipv4 as integers and returns them as such. Their documentation is lacking but
    # research suggests they follow https://tools.ietf.org/html/rfc1700 and big-endian int can be assumed.

    local_ipv4 = nc.get("netconn_local_ipv4", "")
    if local_ipv4:
        local_ipv4 = socket.inet_ntoa(struct.pack("!i", local_ipv4))
    local_ipv6 = nc.get("netconn_local_ipv6", "")
    if local_ipv6:
        local_ipv6 = ipaddress.ip_address(int(local_ipv6, 16))
    local = f"from {local_ipv4}{local_ipv6}:{nc.get('netconn_local_port')}"

    proxy_ipv4 = nc.get("netconn_proxy_ipv4", "")
    if proxy_ipv4:
        proxy_ipv4 = socket.inet_ntoa(struct.pack("!i", proxy_ipv4))
    proxy_ipv6 = nc.get("netconn_proxy_ipv6", "")
    if proxy_ipv6:
        proxy_ipv6 = ipaddress.ip_address(int(proxy_ipv6, 16))
    proxy = ""
    if proxy_ipv6 or proxy_ipv4:
        proxy = f" proxied via {proxy_ipv4}{proxy_ipv6}:{nc.get('netconn_proxy_port')}:{nc.get('netconn_remote_port')}"

    remote_ipv4 = nc.get("netconn_remote_ipv4", "")
    if remote_ipv4:
        remote_ipv4 = socket.inet_ntoa(struct.pack("!i", remote_ipv4))
    remote_ipv6 = nc.get("netconn_remote_ipv6", "")
    if remote_ipv6:
        remote_ipv6 = ipaddress.ip_address(int(remote_ipv6, 16))
    remote = ""
    if remote_ipv4 or remote_ipv6:
        remote = f"{remote_ipv4}{remote_ipv6}:{nc.get('netconn_remote_port')} "

    domain = f"domain={nc.get('netconn_domain')}" if nc.get("netconn_domain") else ""
    return f" @{as_configured_timezone(nc.get('event_timestamp'))}: {action} {direction} {protocol} {local}{proxy} to {remote}{domain}"


def print_netconns(
    p: Union[Process, Dict], raw_print=False, start_time: datetime.datetime = None, end_time: datetime.datetime = None
):
    """Print network connection events. Both netconn and netconn_proxy events.

    action is one or more of: ACTION_CONNECTION_CREATE, ACTION_CONNECTION_CLOSE,
     ACTION_CONNECTION_ESTABLISHED, ACTION_CONNECTION_CREATE_FAILED, ACTION_CONNECTION_LISTEN
    """
    print("------ NETCONNS ------")
    for nc in get_events_by_type(p, ["netconn", "netconn_proxy"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(nc)
            continue
        print(format_netconn(nc))
    print()


def format_regmod(rm: Union[Event, Dict]):
    """Format regmod event into single line."""
    actions = []
    for a in rm.get("regmod_action"):
        if a.startswith("ACTION_"):
            actions.append(a[len("ACTION_") :])
        else:
            actions.append(a)
    # how could the list they return ever be greater than 1 with un-corrupted data?
    action = actions[0]  # if len(actions) > 1 else actions
    if "CREATE" in action:
        action = "Created"
    elif action == "WRITE_VALUE":
        action = "Modified"
    return f" @{as_configured_timezone(rm.get('event_timestamp'))}: {action}: {rm.get('regmod_name')}"


def print_regmods(
    p: Union[Process, Dict], raw_print=False, start_time: datetime.datetime = None, end_time: datetime.datetime = None
):
    """Print registry modifications.

    Actions: ACTION_INVALID, ACTION_CREATE_KEY, ACTION_WRITE_VALUE, ACTION_DELETE_KEY, ACTION_DELETE_VALUE,
        ACTION_RENAME_KEY, ACTION_RESTORE_KEY, ACTION_REPLACE_KEY, ACTION_SET_SECURITY
    """
    print("------ REGMODS ------")
    for rm in get_events_by_type(p, ["regmod"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(rm)
            continue
        print(format_regmod(rm))
    print()


def format_scriptload(sl: Union[Event, Dict]):
    """Format scriptload event into single line."""
    if sl.get("event_type") == "scriptload":
        pub_state = ",".join([state[len("FILE_SIGNATURE_") :] for state in sl.get("scriptload_publisher_state", [])])
        return f" @{as_configured_timezone(sl.get('event_timestamp'))}: {sl.get('scriptload_name')} , sha256={sl.get('scriptload_sha256')} - {pub_state}"

    if sl.get("event_type") == "fileless_scriptload":
        return f" @{as_configured_timezone(sl.get('event_timestamp'))}: {sl.get('fileless_scriptload_cmdline')}"


def print_scriptloads(
    p: Union[Process, Dict], raw_print=False, start_time: datetime.datetime = None, end_time: datetime.datetime = None
):
    """Print scriptloads and fileless scriptloads."""
    print("------ SCRIPTLOADS ------")
    for sl in get_events_by_type(p, ["scriptload"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(sl)
            continue
        print(format_scriptload(sl))
    print()
    print("------ FILELESS SCRIPTLOADS ------")
    for sl in get_events_by_type(p, ["fileless_scriptload"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(sl)
            continue
        print(format_scriptload(sl))
    print()


def format_modload(ml: Union[Event, Dict]):
    """Format modload into single line."""
    # "for now can only be: ACTION_LOAD_MODULE"
    # action = "Loaded" if ml.modload_action == 'ACTION_LOAD_MODULE' else ml.modload_action
    ml_pub_state_summary = (
        "_".join([state[len("FILE_SIGNATURE_STATE") + 1 :] for state in ml.get("modload_publisher_state", [])]) or ""
    )
    return f" @{as_configured_timezone(ml.get('event_timestamp'))}: {ml.get('modload_name')} , md5:{ml.get('modload_md5')} - {ml.get('modload_publisher')}: {ml_pub_state_summary}"


def print_modloads(
    p: Union[Process, Dict], raw_print=False, start_time: datetime.datetime = None, end_time: datetime.datetime = None
):
    """Print modual/library loads."""
    print("------ MODLOADS ------")
    for ml in get_events_by_type(p, ["modload"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(ml)
            continue
        print(format_modload(ml))
    print()


def format_crossproc(cp: Union[Event, Dict]):
    """Format crossproc into single line."""
    actions = [a[len("ACTION_") :] for a in cp.get("crossproc_action", [])]
    if len(actions) == 1:
        actions = actions[0]
    else:
        actions = ",".join(actions)
    inverse_target = "from" if cp.get("crossproc_target") is True else "to"
    direction = "<-" if cp.get("crossproc_target") is True else "->"
    proc_guid_direction = f"{cp.get('process_guid')} {direction} {cp.get('crossproc_process_guid')}"
    return f" @{as_configured_timezone(cp.get('event_timestamp'))}: {actions} {inverse_target} {cp.get('crossproc_name')} ({cp.get('crossproc_sha256')}) | {proc_guid_direction}"


def print_crossprocs(
    p: Union[Process, Dict], raw_print=False, start_time: datetime.datetime = None, end_time: datetime.datetime = None
):
    """Print Cross Process activity.

    Actions: ACTION_DUP_PROCESS_HANDLE, ACTION_OPEN_THREAD_HANDLE, ACTION_DUP_THREAD_HANDLE,
        ACTION_CREATE_REMOTE_THREAD, ACTION_API_CALL
    """
    print("------ CROSSPROCS ------")
    for cp in get_events_by_type(p, ["crossproc"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(cp)
            continue
        print(format_crossproc(cp))
    print()


def format_childproc(cp: Union[Event, Dict]):
    """Format childproc event into single line."""
    return f" @{as_configured_timezone(cp.get('event_timestamp'))}: {cp.get('childproc_cmdline')}  - {cp.get('childproc_process_guid')}"


def print_childprocs(
    p: Union[Process, Dict], raw_print=False, start_time: datetime.datetime = None, end_time: datetime.datetime = None
):
    """Print child process events."""
    print("------ CHILDPROCS ------")
    for cp in get_events_by_type(p, ["childproc"], start_time=start_time, end_time=end_time):
        if raw_print:
            print(cp)
            continue
        print(format_childproc(cp))
    print()


def format_event_data(event_data: Dict):
    assert "event_type" in event_data
    if "netconn" in event_data["event_type"]:
        return format_netconn(event_data)
    if event_data["event_type"] == "filemod":
        return format_filemod(event_data)
    if event_data["event_type"] == "modload":
        return format_modload(event_data)
    if "scriptload" in event_data["event_type"]:
        return format_scriptload(event_data)
    if event_data["event_type"] == "childproc":
        return format_childproc(event_data)
    if event_data["event_type"] == "crossproc":
        return format_crossproc(event_data)
    if event_data["event_type"] == "regmod":
        return format_regmod(event_data)
    LOGGER.warning(f"unknown event of type: {event_data['event_type']}")


def inspect_process_tree(
    proc: Process,
    info=False,
    filemods=False,
    netconns=False,
    regmods=False,
    modloads=False,
    crossprocs=False,
    children=False,
    scriptloads=False,
    max_depth=0,
    depth=0,
    start_time: datetime.datetime = None,
    end_time: datetime.datetime = None,
    **kwargs,
):
    """Walk down the execution chain and print inspection points."""
    if max_depth and depth > max_depth:
        return

    if depth == 0:
        print_ancestry(proc)
        print_process_tree(proc, start_time=start_time, end_time=end_time)

    process_name = get_os_independent_filepath(proc.get("process_name", "None")).name
    print(f"\n+ {process_name} - {proc.process_guid}")
    if info:
        print_process_info(proc, **kwargs)
    if filemods:
        print_filemods(proc, start_time=start_time, end_time=end_time, **kwargs)
    if netconns:
        print_netconns(proc, start_time=start_time, end_time=end_time, **kwargs)
    if regmods:
        print_regmods(proc, start_time=start_time, end_time=end_time, **kwargs)
    if modloads:
        print_modloads(proc, start_time=start_time, end_time=end_time, **kwargs)
    if crossprocs:
        print_crossprocs(proc, start_time=start_time, end_time=end_time, **kwargs)
    if children:
        print_childprocs(proc, start_time=start_time, end_time=end_time, **kwargs)
    if scriptloads:
        print_scriptloads(proc, start_time=start_time, end_time=end_time, **kwargs)

    try:
        for child in proc.children:
            try:
                inspect_process_tree(
                    child,
                    info=info,
                    filemods=filemods,
                    netconns=netconns,
                    regmods=regmods,
                    modloads=modloads,
                    crossprocs=crossprocs,
                    children=children,
                    scriptloads=scriptloads,
                    max_depth=max_depth,
                    depth=depth + 1,
                    start_time=start_time,
                    end_time=end_time,
                    **kwargs,
                )
            except RecursionError:
                LOGGER.warning(f"hit RecursionError inspecting process tree.")
                break
    except ObjectNotFoundError as e:
        LOGGER.warning(f"got object not found error for child proc: {e}")



"""Functions that work with query related Carbon Black APIs.
"""


import time
import datetime
import logging

from typing import Union, Dict, List

# NOTE: boil everything down to CbPSCBaseAPI where possible
# so "enterprise standard" will work wherever possible?
# from cbapi.psc.rest_api import CbPSCBaseAPI
from cbapi.psc.threathunter import CbThreatHunterAPI, Process
from cbapi.psc.threathunter.models import AsyncProcessQuery
from cbapi.errors import ServerError, ClientError, ObjectNotFoundError

LOGGER = logging.getLogger("cbinterface.psc.query")


def create_event_search(
    p: Union[Process, Dict],
    search_data: Dict = {},
    criteria: Dict = {},
    fields: List = ["*"],
    query: str = None,
    time_range: Dict = {},
    rows=1000,
    start: int = 0,
    sort: Dict = [{"field": "event_timestamp", "order": "asc"}],
) -> Dict:
    """Perform an event search.

    NOTE: If p is a dictionary, an instance of CBC API must be passed as "_cb".

    Without anything specified, the default is to return ALL events for the process.
    """
    # NOTE that this one is not job based search.

    cb = p.get("_cb")
    url = f"/api/investigate/v2/orgs/{cb.credentials.org_key}/events/{p.get('process_guid')}/_search"

    if not search_data:
        if not query:
            query = f"process_guid:{p.get('process_guid')}"
        search_data = {
            "criteria": criteria,
            "fields": fields,
            "query": query,
            "rows": rows,
            "start": start,
            "sort": sort,
        }
        if time_range:
            # "time_range" = { "end": "2020-01-27T18:34:04Z", "start": "2020-01-18T18:34:04Z"}
            search_data["time_range"] = time_range

    try:
        result = cb.post_object(url, search_data)
        return result.json()
    except ServerError as e:
        LOGGER.error(f"Caught ServerError searching events: {e}")
        return False
    except ClientError as e:
        LOGGER.warning(f"got ClientError searching events: {e}")
        return False
    except ValueError:
        LOGGER.warning(f"got unexpected {result}")
        return False


def event_search_complete(cb: CbThreatHunterAPI, job_id):
    """Return true when a search is complete."""
    url = f"/api/investigate/v1/orgs/{cb.credentials.org_key}/enriched_events/search_jobs/{job_id}"
    result = cb.get_object(url)
    if result["completed"] == result["contacted"]:
        return True
    return False


def get_event_search_results(cb: CbThreatHunterAPI, job_id) -> Dict:
    """Return any results of an event search."""
    url = f"/api/investigate/v2/orgs/{cb.credentials.org_key}/enriched_events/search_jobs/{job_id}/results"
    try:
        while not event_search_complete(cb, job_id):
            time.sleep(0.1)
    except Exception as e:
        LOGGER.error(f"got exception waiting for event search to complete: {e}")
        return None
    try:
        return cb.get_object(url)
    except Exception as e:
        LOGGER.error("could not get results: {e}")
        return None


def yield_events(
    p: Process,
    search_data: Dict = {},
    criteria: Dict = {},
    query: str = None,
    rows=1000,
    start: int = 0,
    max_results: int = None,  # limit results returned
    start_time: datetime.datetime = None,
    end_time: datetime.datetime = None,
) -> Dict:
    """Yield Process Events resulting from Event search."""

    time_range = {}
    if start_time:
        time_range["start"] = start_time.isoformat()
    if end_time:
        time_range["end"] = end_time.isoformat()

    position = start
    still_querying = True
    while still_querying:
        result = create_event_search(
            p,
            search_data=search_data,
            criteria=criteria,
            query=query,
            time_range=time_range,
            rows=rows,
            start=position,
        )
        if not result:
            return result
        LOGGER.debug(
            f"got result (minus events): {[f'{key}={result[key]}' for key in result.keys() if key !='results']}"
        )
        if max_results and position + rows > max_results:
            # get however many rows that may result in max_results
            rows = max_results - position

        total_results = result["num_available"]
        if total_results != result["num_found"]:
            LOGGER.debug(f"not all events are available.")
        results = result.get("results", [])

        LOGGER.debug(f"got {len(results)+position} out of {total_results} total events.")
        for item in results:
            yield item
            position += 1
            if max_results and position >= max_results:
                still_querying = False
                break
        if position >= total_results:
            if result.get("processed_segments") != result.get("total_segments"):
                LOGGER.warning(
                    f"got all available events but CBC reports that all process segments have not been processed."
                )
            still_querying = False
            break


def get_process_search_jobs(cb):
    url = f"/api/investigate/v1/orgs/{cb.credentials.org_key}/processes/search_jobs"
    return cb.get_object(url)


def get_process_search_status(cb, job_id):
    url = f"/api/investigate/v1/orgs/{cb.credentials.org_key}/processes/search_jobs/{job_id}"
    return cb.get_object(url)


def get_process_search_results(cb, job_id):
    url = f"/api/investigate/v2/orgs/{cb.credentials.org_key}/processes/search_jobs/{job_id}/results"
    return cb.get_object(url)


def cancel_process_search(cb, job_id):
    url = f"/api/investigate/v1/orgs/{cb.credentials.org_key}/processes/search_jobs/{job_id}"
    return cb.delete_object(url)


def is_valid_process_query(query: AsyncProcessQuery) -> bool:
    """Custom query validation.

    Args:
        query: Cb threathunter query
    Returns:
        True if a valid query, else False.
    """
    if not query._doc_class.validation_url:
        LOGGER.debug("Unexpectedly, class has no validation url. hmm...")
        return True

    url = query._doc_class.validation_url.format(query._cb.credentials.org_key)

    args = query._get_query_parameters()
    if args.get("query", False):
        args["q"] = args["query"]
    # v2 search sort key does not work with v1 validation
    args.pop("sort", None)
    LOGGER.debug(f"attempting to validate query with args: {args}")

    validated = query._cb.get_object(url, query_parameters=args)
    if not validated.get("valid"):
        LOGGER.error(f'Invalud query {validated["invalid_message"]}')
        return False
    return True


def is_valid_process_query_string(cb: CbThreatHunterAPI, query: str) -> bool:
    """
    Validates a process query string is valid for PSC.

    Args:
        cb: Cb PSC connection object
        query (str): The query.
    Returns:
        True or False
    """
    args = {"q": query}
    url = f"/api/investigate/v1/orgs/{cb.credentials.org_key}/processes/search_validation"
    validated = cb.get_object(url, query_parameters=args)
    if not validated.get("valid"):
        return False
    return True


def convert_from_legacy_query(cb: CbThreatHunterAPI, query: str) -> str:
    """
    Converts a legacy CB Response query to a ThreatHunter query.

    Args:
        cb: Cb PSC connection object
        query (str): The query to convert.
    Returns:
        str: The converted query.
    """
    args = {"query": query}
    resp = cb.post_object("/threathunter/feedmgr/v2/query/translate", args)
    if resp.status_code != 200:
        LOGGER.error(f"got {resp.status_code} attempting query conversion")
        return False
    resp = resp.json()
    return resp.get("query")


def make_process_query(
    cb: CbThreatHunterAPI,
    query: str,
    start_time: datetime.datetime = None,
    last_time: datetime.datetime = None,
    raise_exceptions=True,
    validate_query=False,
) -> AsyncProcessQuery:
    """Query the CbThreatHunterAPI environment and interface results.

    Args:
        cb: A CbThreatHunterAPI object to use
        query: The process query
        start_time: Set the process start time (UTC).
        last_time: Set the process last time (UTC). Only processes with a start
        time that falls before this last_time.
        raise_exceptions: Let any exceptions raise up (library use)
        validate_query: If True, validate the query before attempting to use it.
    Returns: AsyncProcessQuery or empty list.
    """

    LOGGER.debug(f"buiding query: {query} between '{start_time}' and '{last_time}'")
    processes = []
    try:
        processes = cb.select(Process).where(query)
        if validate_query and not is_valid_process_query(processes):
            LOGGER.info(f"For help, refer to {cb.url}/#userGuideLocation=search-guide/investigate-th&fullscreen")
            LOGGER.info(f"Is this a legacy query? ... Attempting to convert to PSC query ...")
            converted_query = convert_from_legacy_query(cb, query)
            if not converted_query:
                LOGGER.info(f"failed to convert to PSC query... 🤡 your query is jacked up.")
                return []
            if is_valid_process_query_string(cb, converted_query):
                LOGGER.info("successfully converted and validated the query you supplied to a PSC query 👍, see below.")
                LOGGER.info(f"👇👇 try again with the following query 👇👇 - also, hint, single quotes are your friend. ")
                LOGGER.info(f"query: '{converted_query}'")
            return []
        if start_time or last_time:
            start_time = start_time.isoformat() if start_time else "*"
            end_time = last_time.isoformat() if last_time else "*"
            processes = processes.where(f"process_start_time:[{start_time} TO {end_time}]")
        LOGGER.info(f"got {len(processes)} process results.")
    except Exception as e:
        if raise_exceptions:
            raise (e)
        LOGGER.error(f"unexpected exception: {e}")

    return processes


def print_facet_histogram(processes: AsyncProcessQuery):
    """Print facets"""
    # NOTE, this is a custom implementations. TODO, look at using the built in
    # API methods: https://developer.carbonblack.com/reference/carbon-black-cloud/cb-threathunter/latest/process-search-v2/#start-a-process-facet-job
    # Also, NOTE that this table lists fields that support faceting via the built in method, children is not one of them:
    # https://developer.carbonblack.com/reference/cb-threathunter/latest/process-search-fields/
    from cbinterface.helpers import create_histogram_string, get_os_independent_filepath

    fields = [
        "parent_name",
        "process_name",
        "process_reputation",
        "process_username",
        "process_sha256",
        "device_name",
        "device_os",
    ]
    path_fields = ["parent_name", "process_name"]
    processes = list(processes)
    facet_dict = {}
    for field_name in fields:
        facet_dict[field_name] = {}
        for proc in processes:
            value = proc.get(field_name, "None")
            if isinstance(value, list):
                if len(value) > 1:
                    LOGGER.info(f"condensing {value} to {value[0]}")
                value = value[0]
            elif field_name in path_fields:
                file_path = get_os_independent_filepath(value)
                file_name = file_path.name
                value = file_name
            if value not in facet_dict[field_name]:
                facet_dict[field_name][value] = 1
            else:
                facet_dict[field_name][value] += 1

    # special case for "children"
    try:
        facet_dict["childproc_name"] = {}
        depth = 0
        for proc in processes:
            if proc.childproc_count < 1:
                continue
            children = proc.summary.children or []
            for cp in children:
                process_path = get_os_independent_filepath(cp.get("process_name"))
                process_name = process_path.name
                if process_name not in facet_dict["childproc_name"]:
                    facet_dict["childproc_name"][process_name] = 1
                else:
                    facet_dict["childproc_name"][process_name] += 1
    except Exception as e:
        LOGGER.warning(f"problem enumerating child process names: {e}")

    print("\n------------------------- FACET HISTOGRAMS -------------------------")
    for field_name, facets in facet_dict.items():
        print(f"\n\t{field_name} results: {len(facets.keys())}")
        print("\t--------------------------------")
        print(create_histogram_string(facets))
    return


def print_facet_histogram_v2(
    cb: CbThreatHunterAPI,
    query: str,
    start_time: datetime.datetime = None,
    end_time: datetime.datetime = None,
    return_string=False,
):
    """Get query facet results from the CbAPI enriched events facets."""

    # NOTE: no support for childproc facets with this built-in

    from cbinterface.helpers import get_os_independent_filepath

    post_data = {}
    post_data["query"] = query
    fields = [
        "parent_name",
        "process_name",
        "process_reputation",
        "process_username",
        "process_sha256",
        "device_name",
        "device_os",
    ]
    path_fields = ["parent_name", "process_name"]
    post_data["terms"] = {"fields": fields}
    post_data["time_range"] = {}
    if start_time:
        post_data["time_range"]["start"] = start_time.isoformat()
    if end_time:
        post_data["time_range"]["end"] = end_time.isoformat()

    # TODO handle status_code!=200 and response is not json for both requests

    uri = f"/api/investigate/v2/orgs/{cb.credentials.org_key}/processes/facet_jobs"
    job_id = cb.post_object(uri, post_data).json().get("job_id", None)
    if not job_id:
        LOGGER.error(f"failed to get facet job.")
        return False

    uri = f"/api/investigate/v2/orgs/{cb.credentials.org_key}/processes/facet_jobs/{job_id}/results"
    facet_data = cb.get_object(uri)

    txt = "\n------------------------- FACET HISTOGRAMS -------------------------\n"
    total = facet_data["num_found"]
    for facets in facet_data["terms"]:
        field_name = facets["field"]
        txt += f"\n\t{field_name} results: {len(facets['values'])}\n"
        txt += "\t--------------------------------\n"
        for entry in facets["values"]:
            entry_name = entry["name"]
            if field_name in path_fields and len(entry_name) > 55:
                file_path = get_os_independent_filepath(entry_name)
                file_name = file_path.name
                file_path = entry_name[: len(entry_name) - len(file_name)]
                file_path = file_path[: 40 - len(file_name)]
                entry_name = f"{file_path}...{file_name}"
            bar_value = int(((entry["total"] / total) * 100) / 2)
            txt += "%30s: %5s %5s%% %s\n" % (
                entry_name,
                entry["total"],
                int(entry["total"] / total * 100),
                "\u25A0" * bar_value,
            )

    txt += "\n"
    if return_string:
        return txt
    print(txt)
    return



"""Helper functions for common actions.
"""

import os
import re
import sys
import signal
import logging
import datetime

from pathlib import PureWindowsPath, PurePosixPath
from typing import Union

from dateutil import tz
from dateutil.parser import isoparse
from dateutil.zoneinfo import get_zonefile_instance

LOGGER = logging.getLogger("cbinterface.helpers")

UUID_REGEX = re.compile(r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", re.I)
PSC_GUID_REGEX = re.compile(r"[0-9A-Z]{8}-[0-9a-f]{8}-[0-9a-f]{8}-[0-9a-f]{8}-[0-9a-f]{15}", re.I)


def is_uuid(uuid: str):
    """Returns True if the given string matches the UUID pattern."""
    return UUID_REGEX.match(uuid)


def is_psc_guid(guid: str):
    """Returns True if the given string matches the format of a PSC GUID."""
    return PSC_GUID_REGEX.match(guid)


def input_with_timeout(prompt, default=None, timeout=30, stderr=True):
    """Wait up to timeout for user input"""

    def _log_and_exit(signum, frame):
        if stderr:
            sys.stderr.write("\n")
        else:
            sys.stdout.write("\n")
        LOGGER.error("Timeout reached waiting for input.")
        sys.exit()

    signal.signal(signal.SIGALRM, _log_and_exit)
    signal.alarm(timeout)
    if stderr:
        sys.stderr.write(prompt)
    else:
        sys.stdout.write(prompt)
    answer = input() or default
    signal.alarm(0)
    return answer


def clean_exit(signal, frame):
    print()
    LOGGER.info(f"caught KeyboardInterrupt. exiting.")
    sys.exit(0)


def get_os_independent_filepath(unknown_os_file_path: str) -> Union[PureWindowsPath, PurePosixPath]:
    """Return a proper os filepath object."""
    filepath = PureWindowsPath(unknown_os_file_path)
    nixfilepath = PurePosixPath(unknown_os_file_path)
    if len(nixfilepath.parts) > len(filepath.parts):
        filepath = nixfilepath
    return filepath


def as_configured_timezone(timestamp: Union[datetime.datetime, str], apply_time_format="%Y-%m-%d %H:%M:%S.%f%z") -> str:
    """Convert timestamp to the configured time zone."""
    from cbinterface.config import get_timezone

    if not timestamp:
        return timestamp

    if isinstance(timestamp, str):
        # psc
        try:
            timestamp = isoparse(timestamp)
        except ValueError:
            return ""

    if isinstance(timestamp, str):
        # legacy format
        try:
            timestamp = datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S.%f")
        except ValueError:
            LOGGER.warning(f"ValueError trying to parse '{timestamp}' with '%Y-%m-%d %H:%M:%S.%f'")
            return timestamp

    # the timestamps from CbR are not timezone aware, but they are GMT.
    timestamp = timestamp.replace(tzinfo=tz.gettz("GMT")) if timestamp.tzinfo is None else timestamp
    return timestamp.astimezone(get_timezone()).strftime(apply_time_format)


def utc_offset_to_potential_tz_names(utc_offset: datetime.timedelta):
    potential_zones = []
    utc_now = datetime.datetime.utcnow()
    for zone in list(get_zonefile_instance().zones):
        dt = utc_now.astimezone(tz.gettz(zone))
        if dt.utcoffset() == utc_offset:
            potential_zones.append(zone)

    return list(sorted(potential_zones))


def create_histogram_string(data: dict) -> str:
    """A convenience function that creates a graph in the form of a string.
    Args:
        data: A dictionary, where the values are integers representing a count of the keys.
    Returns:
        A graph in string form, pre-formatted for raw printing.
    """

    assert isinstance(data, dict)
    for key in data.keys():
        assert isinstance(data[key], int)

    total_results = sum([value for value in data.values()])
    txt = ""

    # order keys for printing in order (purly ascetics)
    ordered_keys = sorted(data, key=lambda k: data[k])
    results = []

    # longest_key used to calculate how many white spaces should be printed
    # to make the graph columns line up with each other
    longest_key = 0
    for key in ordered_keys:
        value = data[key]
        longest_key = len(key) if len(key) > longest_key else longest_key
        # IMPOSING LIMITATION: truncating keys to 95 chars, keeping longest key 5 chars longer
        longest_key = 100 if longest_key > 100 else longest_key
        percent = value / total_results * 100
        results.append((key[:95], value, percent, "\u25A0" * (int(percent / 2))))

    # two for loops are ugly, but allowed us to count the longest_key -
    # so we loop through again to print the text
    for r in results:
        txt += "%s%s: %5s - %5s%% %s\n" % (
            int(longest_key - len(r[0])) * " ",
            r[0],
            r[1],
            str(r[2])[:4],
            "\u25A0" * (int(r[2] / 2)),
        )
    return txt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

# assumes we're at the repo root
from cbinterface import __version__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Load requirements
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = [line.strip() for line in f.readlines()]

setup(
    name='cbinterface',
    version=__version__,
    description="Command line tool for interfacing with carbonblack environments (PSC & Response) for analysis/investigations and live response playbooks.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/ace-ecosystem/cbinterface2",
    author='Sean McFeely',
    author_email='mcfeelynaes@gmail.com',
    license='Apache-2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        "Intended Audience :: Information Technology",
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
    ],
    keywords='Carbon Black,carbonblack',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    scripts=['bin/cbinterface'],
)



"""Functions that work with Carbon Black Sensors.
"""

import datetime
import logging

from typing import Union

from cbapi.psc import Device
from cbapi.psc.threathunter import CbThreatHunterAPI
from cbapi.psc.devices_query import DeviceSearchQuery

from cbinterface.helpers import as_configured_timezone

LOGGER = logging.getLogger("cbinterface.psc.device")


def is_device_online(d: Device) -> bool:
    """Return True if the device has check in within the last 15 minutes."""
    elapsed_time = time_since_checkin(d)
    LOGGER.debug(f"elapsed time since device checked in: {elapsed_time}")
    if elapsed_time < datetime.timedelta(minutes=15):
        # guesstimation. Cb documentation is lacking.
        return True
    return False


def make_device_query(cb: CbThreatHunterAPI, device_query: str) -> DeviceSearchQuery:
    """Construct a DeviceSearchQuery object."""
    try:
        if ":" not in device_query:
            LOGGER.info("No field specification passed. Use 'FIELDS' for help.")
        devices = cb.select(Device).where(device_query)
    except ValueError as e:
        LOGGER.error(f"{e}")
        return False
    LOGGER.info(f"got {len(devices)} device results.")
    return devices


def find_device_by_hostname(cb: CbThreatHunterAPI, name: str) -> Device:
    """Find a Device by name."""
    devices = make_device_query(cb, f"name:{name}")
    if len(devices) == 1:
        return devices.first()
    elif len(devices) > 1:
        LOGGER.warning(f"{len(devices)} devices with name={name}")
        return None
    return None


def time_since_checkin(device: Device, refresh=True) -> datetime.timedelta:
    """Return the time since last device checkin."""
    from dateutil import tz
    from dateutil.parser import isoparse

    if not device.get("last_contact_time"):
        return None

    if refresh:
        device.refresh()
    now = datetime.datetime.utcnow().replace(tzinfo=tz.UTC)
    return now - isoparse(device.last_contact_time)


def device_info(device: Device, refresh=False):
    """Print device info."""
    if refresh:
        device.refresh()
    text = "\n"
    text += "-------------------------------------------------------------------------------\n"
    text += f"\tAD Group ID: {device.ad_group_id}\n"
    text += f"\tCurrent Policy Name: {device.current_sensor_policy_name}\n"
    text += f"\tDeployment Type: {device.deployment_type}\n"
    text += f"\tDevice ID: {device.id}\n"
    text += f"\tDevice Name: {device.name}\n"
    text += f"\tDevice MAC address: {device.mac_address}\n"
    text += f"\tDevice OS: {device.os}\n"
    text += f"\tDevice OS Version: {device.os_version}\n"
    text += f"\tDevice Owner ID: {device.device_owner_id}\n"
    text += f"\tDevice Owner Email: {device.email}\n"
    text += f"\tDevice Owner Name: {device.last_name}, {device.first_name}\n"
    quarantined_message = "🚫 DEVICE QUARANTINED 🚫" if device.quarantined else device.quarantined
    text += f"\tDevice Quarantined: {quarantined_message}\n"
    text += f"\tDevice Registration Time: {as_configured_timezone(device.registered_time)}\n"
    text += f"\tLast Checkin Time: {as_configured_timezone(device.get('last_contact_time'))}\n"
    elapsed_time = time_since_checkin(device, refresh=False)
    online_emotion = "appears online ✅" if elapsed_time < datetime.timedelta(minutes=15) else "likely offline 💤"
    text += "\t " + "\u21B3" + f" Elapsed Time: {elapsed_time} - {online_emotion}\n"
    text += f"\tLast Reported Event Time: {as_configured_timezone(device.last_reported_time)}\n"
    text += f"\tLast External IP: {device.last_external_ip_address}\n"
    text += f"\tLast Internal IP: {device.last_internal_ip_address}\n"
    text += f"\tLast Location: {device.last_location}\n"
    text += f"\tLast Logged In User: {device.login_user_name}\n"
    text += f"\tSensor status: {device.status}\n"
    text += f"\tSensor Version: {device.sensor_version}\n"
    text += "\n"
    return text



"""Configuration related items.
"""

import os
import logging
import glob

from configparser import ConfigParser
from dateutil import tz
from dateutil.zoneinfo import get_zonefile_instance

LOGGER = logging.getLogger("cbinterface.config")

HOME_PATH = os.path.dirname(os.path.abspath(__file__))

user_config_path = os.path.join(os.path.expanduser("~"), ".carbonblack", "cbinterface.ini")
CONFIG_SEARCH_PATHS = [
    "/etc/carbonblack/cbinterface.ini",
    user_config_path,
]

CONFIG = ConfigParser()
CONFIG.read(CONFIG_SEARCH_PATHS)

if not CONFIG.has_section("default"):
    CONFIG.add_section("default")

if not CONFIG.has_option("default", "max_recursive_depth"):
    CONFIG.set("default", "max_recursive_depth", "50")
MAX_RECURSIVE_DEPTH = CONFIG.getint("default", "max_recursive_depth")
# TODO need to actually use max recursive depth and max segments


def save_configuration(config: ConfigParser = CONFIG, save_path=user_config_path):
    """Write config to save_path."""
    if save_path == user_config_path:
        try:
            if not os.path.exists(user_config_path):
                os.mkdir(os.path.join(os.path.expanduser("~"), ".carbonblack"))
        except FileExistsError:
            pass
    try:
        with open(save_path, "w") as fp:
            config.write(fp)
    except FileNotFoundError:
        LOGGER.error(f"part of path does not exist: {save_path}")
    if os.path.exists(save_path):
        LOGGER.info(f"saved configuration to: {save_path}")


def _is_valid_time_zone(zone: str):
    """True if zone is a valid time zone."""
    if zone in list(get_zonefile_instance().zones):
        return True
    return False


# begin cbinterface timezone
def set_timezone(time_zone: str):
    if not _is_valid_time_zone(time_zone):
        LOGGER.error(f"Not a recongnized time zone: {time_zone}")
        return False
    os.environ["CBINTERFACE_TIMEZONE"] = time_zone
    CONFIG.set("default", "time_zone", time_zone)
    return True


def get_timezone():
    return tz.gettz(os.environ.get("CBINTERFACE_TIMEZONE", "GMT"))


if "CBINTERFACE_TIMEZONE" not in os.environ:
    # timestamps from Cb are not timezone aware, but they are GMT.
    set_timezone(CONFIG.get("default", "time_zone", fallback="GMT"))

# begin CbAPI product
def set_default_cbapi_product(env_product: str):
    os.environ["CBINTERFACE_DEFAULT_CBAPI_PRODUCT"] = env_product
    CONFIG.set("default", "cbapi_product", env_product)
    return True


def get_default_cbapi_product():
    return os.environ.get("CBINTERFACE_DEFAULT_CBAPI_PRODUCT", "response")


if "CBINTERFACE_DEFAULT_CBAPI_PRODUCT" not in os.environ:
    set_default_cbapi_product(CONFIG.get("default", "cbapi_product", fallback="response"))

# begin CbAPI profile/environment
def set_default_cbapi_profile(profile: str):
    os.environ["CBINTERFACE_DEFAULT_CBAPI_PROFILE"] = profile
    CONFIG.set("default", "cbapi_profile", profile)
    return True


def get_default_cbapi_profile():
    return os.environ.get("CBINTERFACE_DEFAULT_CBAPI_PROFILE", "default")


if "CBINTERFACE_DEFAULT_CBAPI_PROFILE" not in os.environ:
    set_default_cbapi_profile(CONFIG.get("default", "cbapi_profile", fallback="default"))


def get_playbook_map():
    """Load playbook config file map.

    Returns: dict of playbook name to playbook config path.
    """
    playbook_map = {}
    # package included playbooks
    playbook_paths = glob.glob(f"{HOME_PATH}/playbook_configs/*.ini")
    # configured playbooks
    if CONFIG.has_section("playbooks"):
        playbook_paths.extend(list(CONFIG["playbooks"].values()))

    for playbook_path in playbook_paths:
        playbook_name = playbook_path[playbook_path.rfind("/") + 1 : playbook_path.rfind(".")]
        playbook = ConfigParser()
        try:
            playbook.read(playbook_path)
        except Exception as e:
            LOGGER.error(f"could not load playbook: {e}")
            continue
        if playbook_name in playbook_map:
            LOGGER.error(f"playbook name collision on '{playbook_name}'. skipping this one...")
            continue
        playbook_name = playbook.get("overview", "name", fallback=playbook_name)
        playbook_description = playbook.get("overview", "description", fallback="")
        playbook_map[playbook_name] = {
            "path": playbook_path,
            "name": playbook_name,
            "description": playbook_description,
        }

    return playbook_map



"""Everthing response process related.
"""

import json
import inspect
import logging

from io import StringIO
from datetime import datetime
from contextlib import redirect_stdout
from typing import Generator, List, Dict

from cbapi.response import Process, models
from cbapi.errors import ObjectNotFoundError

from cbinterface.helpers import as_configured_timezone

LOGGER = logging.getLogger("cbinterface.response.process")


def process_to_dict(p: Process, max_segments=None) -> Dict:
    """Get all events for this process."""

    all_segments = p.get_segments()
    if max_segments is None:
        max_segments = len(all_segments)

    p.refresh()
    results = p.original_document
    results["captured_segments"] = {}
    results["all_segments"] = all_segments

    results["process_ancestry"] = StringIO()
    with redirect_stdout(results["process_ancestry"]):
        print_ancestry(p)
    results["process_ancestry"] = results["process_ancestry"].getvalue()

    results["process_tree"] = StringIO()
    with redirect_stdout(results["process_tree"]):
        print_process_tree(p)
    results["process_tree"] = results["process_tree"].getvalue()

    captured_segment_count = 0
    if p.current_segment:
        # if current_segment is set, something specifically targeted this segment
        # and we will ensure it gets captured here
        results["captured_segments"][p.current_segment] = segment_events_to_dict(p)
        captured_segment_count += 1

    for segment in all_segments:
        p.current_segment = segment
        if segment in results["captured_segments"]:
            continue
        if captured_segment_count >= max_segments:
            LOGGER.info(f"hit maximum segment limit exporting process to json for {p.id}")
            break
        results["captured_segments"][segment] = segment_events_to_dict(p)
        captured_segment_count += 1

    return results


def cb_event_to_dict(cbevent: models.CbEvent) -> Dict:
    """Convert a single CbEvent to dict.

    Get the default fields and enumerate some more.
    """

    def _is_jsonable(item):
        try:
            json.dumps(item)
            return True
        except (TypeError, OverflowError):
            return False

    data = {}
    data["event_type"] = cbevent.event_type
    for title in cbevent.stat_titles:
        data[title] = str(getattr(cbevent, title, ""))

    all_members = [member for member in inspect.getmembers(cbevent) if not member[0].startswith("_")]
    members = [member for member in all_members if not inspect.ismethod(member[1]) and not inspect.isclass(member[1])]
    for title, attribute in members:
        if title not in data and _is_jsonable(attribute):
            data[title] = attribute

    return data


def cb_events_to_dict(cbevents: Generator[models.CbEvent, None, None]) -> List[Dict]:
    """Convert a list of CbEvents to a list of dictionaries."""
    return [cb_event_to_dict(event) for event in cbevents]


def segment_events_to_dict(p: Process) -> Dict:
    """Convert current segment CbEvents to dict."""
    if not p.current_segment:
        return {}
    return {
        "filemods": cb_events_to_dict(p.filemods),
        "netconns": cb_events_to_dict(p.netconns),
        "regmods": cb_events_to_dict(p.regmods),
        "modloads": cb_events_to_dict(p.modloads),
        "crossprocs": cb_events_to_dict(p.crossprocs),
        "children": cb_events_to_dict(p.children),
    }


def print_process_info(proc: Process, return_string: bool = False, raw_print=False, header=True):
    """Analyst friendly custom process data format.

    Args:
        proc: CbR Process
        return_string: return string if True, else print it to stdout.
    Returns: string or None
    """

    if not proc._info and raw_print:
        LOGGER.debug(f"retrieving process info.")
        proc.refresh()

    txt = ""
    if header:
        txt += "------ INFO ------\n"
    if raw_print:
        txt = str(proc)
    else:
        status = "Terminated" if proc.terminated else "Running"
        txt += f"  Process GUID: {proc.id}\n"
        txt += f"  Process Name: {proc.process_name}\n"
        txt += f"  Process PID: {proc.process_pid}\n"
        txt += f"  Process MD5: {proc.process_md5}\n"
        txt += f"  Process Path: {proc.path}\n"
        txt += f"  Process Status: {status}\n"
        txt += f"  Command Line: {proc.cmdline}\n"
        txt += f"  Parent Name: {proc.parent_name}\n"
        txt += f"  Parent GUID: {proc.parent_id}\n"
        txt += f"  Hostname: {proc.hostname}\n"
        txt += f"  Username: {proc.username}\n"
        txt += f"  Start Time: {as_configured_timezone(proc.start)}\n"
        try:
            txt += f"  Last Update Time: {as_configured_timezone(proc.last_update)}\n"
        except TypeError:  # should be handled by cbapi
            txt += f"  Last Update Time: None\n"
        txt += f"  Sensor ID: {proc.sensor_id}\n"
        txt += f"  Comms IP: {proc.comms_ip}\n"
        txt += f"  Interface IP: {proc.interface_ip}\n"
        txt += f"  GUI Link: {proc.webui_link}\n"
    if return_string:
        return txt
    txt += "\n"
    print(txt)


def print_ancestry(p: Process):
    """Print the process ancestry for this process."""

    def _print_ancestry_details(p, depth):
        suspressed = " (suppressed) " if p.suppressed_process else " "
        print(f"{'  '*(depth + 1)}{as_configured_timezone(p.start) or '<unknown>'}:  {p.cmdline}{suspressed} | {p.id}")

    print("------ Process Ancestry ------")
    print()
    p.walk_parents(_print_ancestry_details)
    print()


def print_process_tree(p: Process):
    """Print the process tree."""

    def _print_process_tree(p, depth):
        suspressed = "(suppressed) " if p.suppressed_process else ""
        print(f"  {'  '*(depth+1)}{suspressed}{p.cmdline}  | {p.id}")  # - proc_guid={p.id})")

    print("------ Process Execution Tree ------")
    print()
    print(f"  {p.cmdline}  | {p.id}")  # - proc_guid={p.id})")
    p.walk_children(_print_process_tree)
    print()


def inspect_process_tree(
    proc: Process,
    info=False,
    filemods=False,
    netconns=False,
    regmods=False,
    modloads=False,
    crossprocs=False,
    children=False,
    max_depth=0,
    depth=0,
    **kwargs,
):
    """Walk down the execution chain and print inspection points."""
    if max_depth and depth > max_depth:
        return

    if depth == 0:
        print_ancestry(proc)
        print_process_tree(proc)

    print(f"\n+ {proc.process_name} - {proc.id}")
    if info:
        print_process_info(proc, **kwargs)
    if filemods:
        print_filemods(proc, **kwargs)
    if netconns:
        print_netconns(proc, **kwargs)
    if regmods:
        print_regmods(proc, **kwargs)
    if modloads:
        print_modloads(proc, **kwargs)
    if crossprocs:
        print_crossprocs(proc, **kwargs)
    if children:
        print_childprocs(proc, **kwargs)

    for cpevent in proc.children:
        if not cpevent.terminated:
            try:
                proc = cpevent.process
            except ObjectNotFoundError:
                continue
            else:
                inspect_process_tree(
                    proc,
                    info=info,
                    filemods=filemods,
                    netconns=netconns,
                    regmods=regmods,
                    modloads=modloads,
                    crossprocs=crossprocs,
                    children=children,
                    max_depth=max_depth,
                    depth=depth + 1,
                    **kwargs,
                )


def print_filemods(p: Process, current_segment_only: bool = False, raw_print=False, **kwargs):
    """Print file modifications."""

    def _print_filemod_events(filemods):
        for fm in filemods:
            assert isinstance(fm, models.CbFileModEvent)
            if raw_print:
                print(fm)
                continue
            detail_line = f" @{as_configured_timezone(fm.timestamp)}: {fm.type} {fm.path}"
            if fm.filetype != "Unknown":
                detail_line += f" - type:{fm.filetype}"
            if fm.md5:
                detail_line += f" - md5:{fm.md5}"
            print(detail_line)
        print()

    print("------ FILEMODS ------")
    if current_segment_only:
        _print_filemod_events(p.filemods)
        return
    _print_filemod_events(p.all_filemods())
    return


def print_netconns(p: Process, current_segment_only: bool = False, raw_print=False):
    """Print network connection events."""

    def _print_netconn_events(netconns):
        for nc in netconns:
            assert isinstance(nc, models.CbNetConnEvent)
            if raw_print:
                print(nc)
                continue
            detail_line = f" @{as_configured_timezone(nc.timestamp)}: ({nc.direction}) local_ip_port={nc.local_ip}:{nc.local_port}"
            if nc.proxy_ip != "0.0.0.0":
                detail_line += f" proxy_ip_port={nc.proxy_ip}:{nc.proxy_port}"
            detail_line += f" remote_ip_port={nc.remote_ip}:{nc.remote_port} domain={nc.domain}"
            print(detail_line)
        print()

    print("------ NETCONNS ------")
    if current_segment_only:
        _print_netconn_events(p.netconns)
        return
    _print_netconn_events(p.all_netconns())
    return


def print_regmods(p: Process, current_segment_only: bool = False, raw_print=False):
    """Print registry modifications."""

    def _print_regmod_events(regmods):
        for rm in regmods:
            assert isinstance(rm, models.CbRegModEvent)
            if raw_print:
                print(rm)
                continue
            print(f" @{as_configured_timezone(rm.timestamp)}: {rm.type} {rm.path}")
        print()

    print("------ REGMODS ------")
    if current_segment_only:
        _print_regmod_events(p.regmods)
        return
    _print_regmod_events(p.all_regmods())
    return


def print_modloads(p: Process, current_segment_only: bool = False, raw_print=False):
    """Print modual/library loads."""

    def _print_modload_events(modloads):
        for ml in modloads:
            assert isinstance(ml, models.CbModLoadEvent)
            if raw_print:
                print(ml)
                continue
            sig_status = "signed" if ml.is_signed else "unsigned"
            print(f" @{as_configured_timezone(ml.timestamp)}: ({sig_status}) {ml.path} , md5:{ml.md5}")
        print()

    print("------ MODLOADS ------")
    if current_segment_only:
        _print_modload_events(p.modloads)
        return
    _print_modload_events(p.all_modloads())
    return


def print_crossprocs(p: Process, current_segment_only: bool = False, raw_print=False):
    """Print Cross Process activity."""

    def _print_crossproc_events(crossprocs):
        for cp in crossprocs:
            assert isinstance(cp, models.CbCrossProcEvent)
            if raw_print:
                print(cp)
                continue
            print(
                f" @{as_configured_timezone(cp.timestamp)}: {cp.type} | {cp.source_path} -> {cp.target_path} | {cp.source_proc.id} -> {cp.target_proc.id}"
            )
            # print() # extra space seems to be helpful on the eye with these
        print()

    print("------ CROSSPROCS ------")
    if current_segment_only:
        _print_crossproc_events(p.crossprocs)
        return
    _print_crossproc_events(p.all_crossprocs())
    return


def print_childprocs(p: Process, current_segment_only: bool = False, raw_print=False):
    """Print child process events."""

    if p.current_segment is None:
        # avoids server error calling /api/v4/process/{guid}/{segment}/event
        p.current_segment = p.get_segments()[0]

    def _print_childproc_events(childprocs):
        if raw_print:
            for cp in childprocs:
                print(cp)
            return

        # group start/end childproc events together
        organized_childprocs = {}
        for cp in childprocs:
            guid = cp.procguid[: cp.procguid.rfind("-")]
            if guid in organized_childprocs:
                organized_childprocs[guid].append(cp)
            else:
                organized_childprocs[guid] = [cp]

        for cp_guid, cp_events in organized_childprocs.items():
            # there should only be two events, a spawn and terminate event
            # however, don't make assumptions
            spawn = cp_events[0]
            terminate_cp = None
            if len(cp_events) > 1:
                for cp in cp_events:
                    if cp.timestamp < spawn.timestamp:
                        spawn = cp
                    else:
                        terminate_cp = cp

            status = "unknown"
            try:
                if spawn.is_suppressed:
                    status = "suppressed"
                status = "terminated" if spawn.process.terminated else "running"
                if spawn.process.terminated and terminate_cp is None:
                    LOGGER.debug(f"notice: no termination event found. process must have terminated elsewhere?")
            except ObjectNotFoundError:
                LOGGER.debug(f"child process not found. ")

            print(
                f" @{as_configured_timezone(spawn.timestamp)}: ({status}) {spawn.path} md5={spawn.md5} pid={spawn.pid} - {cp_guid}"
            )
        print()

    print("------ CHILDPROCS ------")
    childprocs = []
    if current_segment_only:
        childprocs = p.childprocs
    else:
        childprocs = p.all_childprocs()

    try:
        _print_childproc_events(childprocs)
        """
        for cp in p.children:
            print(f" @{as_configured_timezone(cp.timestamp)}: {cp.path} md5={cp.md5} pid={cp.pid} - {cp.procguid}")
        print()
        """
    except Exception as e:
        LOGGER.error(f"unhandled exception when enumerating childproc events: {e}")
    return
